# Copyright (c) Microsoft Corporation.
# Licensed under the MIT license.


import logging
import queue
import re
from collections import defaultdict
import torch
from torch.utils.tensorboard._pytorch_graph import NodePy, NodePyIO, NodePyOP, GraphPy

CLASSTYPE_KIND = "ClassType"
GETATTR_KIND = "prim::GetAttr"
CAT_KIND = "aten::cat"
LIST_CONSTRUCT_KIND = "prim::ListConstruct"
LIST_UNPACK_KIND = "prim::ListUnpack"
TUPLE_CONSTRUCT_KIND = "prim::TupleConstruct"
TUPLE_UNPACK_KIND = "prim::TupleUnpack"
CONSTANT_KIND = "prim::Constant"

_logger = logging.getLogger(__name__)


def build_module_graph(model, dummy_input):
    return TorchModuleGraph(model, dummy_input)


def build_graph(model, dummy_input, verbose=False):
    g = TorchProtoGraph(model, dummy_input, verbose)
    return g.graph_def, g.stepstats


def parse_traced_name(module_name):
    prefix = "TracedModule["
    suffix = "]"
    if module_name.startswith(prefix) and module_name.endswith(suffix):
        module_name = module_name[len(prefix) : -len(suffix)]
    return module_name


class TorchGraph:
    """
    This class is to extract pytorch model topology graph by tracing
    """

    def __init__(self, model=None, dummy_input=None, traced_model=None):
        """
        Parameters
        ----------
        model : pytorch model
            The model user wants to speedup
        dummy_input : pytorch tensor
            The dummy input for ```jit.trace```, users should put it on right device before pass in
        traced_model : torch._C.torch.jit.TopLevelTracedModule
            An alredy traced model, if traced_model is not None, then TorchGraph will build the graph
            based on this traced model and won't trace the model again.
        """
        assert torch.__version__ >= "1.3.1"
        # check if the input is legal
        if traced_model is not None:
            assert isinstance(
                traced_model, torch.jit.TopLevelTracedModule
            ) or isinstance(traced_model, torch.jit.RecursiveScriptModule)
            self.trace = traced_model
            # it's ok if the graph is already unpacked
            torch._C._jit_pass_inline(self.trace.graph)
        elif model is not None and dummy_input is not None:
            self.bound_model = model
            self._trace(model, dummy_input)
        else:
            raise Exception(
                "Please provide model & dummy_input or the traced_model as inputs"
            )

    def _trace(self, model, dummy_input):
        training = model.training
        model.eval()
        kw_args = {}
        if torch.__version__ >= "1.6.0":
            # only pytorch with version greater than 1.6.0 has the strict option
            kw_args["strict"] = False
        self.trace = torch.jit.trace(model, dummy_input, **kw_args)
        torch._C._jit_pass_inline(self.trace.graph)
        model.train(training)


class TorchProtoGraph(TorchGraph):
    """
    Generates model graph for pytorch models in protobuf, this implementation
    is borrowed from pytorch v1.4.0, and fixed following issues:
    https://github.com/pytorch/pytorch/issues/33691
    https://github.com/pytorch/pytorch/issues/33670

    """

    def __init__(self, model, dummy_input, verbose=False):
        super().__init__(model, dummy_input)

        from tensorboard.compat.proto.config_pb2 import RunMetadata
        from tensorboard.compat.proto.graph_pb2 import GraphDef
        from tensorboard.compat.proto.step_stats_pb2 import StepStats, DeviceStepStats
        from tensorboard.compat.proto.versions_pb2 import VersionDef

        list_of_nodes = self.parse(self.trace.graph, self.trace, dummy_input)
        if verbose:
            print(self.trace.graph)
        self.stepstats = RunMetadata(
            step_stats=StepStats(dev_stats=[DeviceStepStats(device="/device:CPU:0")])
        )
        self.graph_def = GraphDef(node=list_of_nodes, versions=VersionDef(producer=22))

    def parse(self, graph, trace, args=None, omit_useless_nodes=True):
        """This method parses an optimized PyTorch model graph and produces
        a list of nodes and node stats for eventual conversion to TensorBoard
        protobuf format.

        Args:
        graph (PyTorch module): The model graph to be parsed.
        trace (PyTorch JIT TracedModule): The model trace to be parsed.
        args (tuple): input tensor[s] for the model.
        omit_useless_nodes (boolean): Whether to remove nodes from the graph.
        """
        nodes_py = GraphPy()
        for node in graph.inputs():
            if omit_useless_nodes:
                if (
                    not node.uses()
                ):  # number of user of the node (= number of outputs/ fanout)
                    continue

            if node.type().kind() != CLASSTYPE_KIND:
                nodes_py.append(NodePyIO(node, "input"))

        attr_to_scope = dict()

        def node_to_name(d):
            return str(d).split(":")[0].strip()

        for node in graph.nodes():
            if node.kind() == GETATTR_KIND:
                attr_name = node.s("name")
                node_name = node_to_name(node)
                parent = node.input().node()
                # If the parent node is not the top-level "self" node
                if parent.kind() == GETATTR_KIND:
                    parent_scope = attr_to_scope[node_to_name(parent)]
                    attr_scope = parent_scope.split("/")[-1]
                    attr_to_scope[node_name] = "{}/{}.{}".format(
                        parent_scope, attr_scope, attr_name
                    )
                else:
                    attr_to_scope[node_name] = "__module.{}".format(attr_name)
                # We don't need classtype nodes; scope will provide this information
                if node.output().type().kind() != CLASSTYPE_KIND:
                    node_py = NodePyOP(node)
                    node_py.scopeName = attr_to_scope[node_name]
                    nodes_py.append(node_py)
            else:
                nodes_py.append(NodePyOP(node))

        # Create sink nodes for output ops
        for i, node in enumerate(graph.outputs()):
            node_py = NodePyIO(node, "output")
            node_py.debugName = "output.{}".format(i + 1)
            node_py.inputs = [node.debugName()]
            nodes_py.append(node_py)

        alias_to_name = dict()
        base_name = parse_traced_name(trace._name)
        for name, module in trace.named_modules(prefix="__module"):
            mod_name = parse_traced_name(module._name)
            attr_name = name.split(".")[-1]
            alias_to_name[name] = "{}[{}]".format(mod_name, attr_name)

        for node in nodes_py.nodes_op:
            module_aliases = node.scopeName.split("/")[-1].split(".")
            module_name = ""
            for i, alias in enumerate(module_aliases):
                if i == 0:
                    module_name = alias
                    node.scopeName = base_name
                else:
                    module_name += "." + alias
                    node.scopeName += "/" + (
                        alias_to_name[module_name]
                        if module_name in alias_to_name
                        else alias
                    )

        nodes_py.populate_namespace_from_OP_to_IO()
        return nodes_py.to_proto()


class NodePyGroup(NodePy):
    """
    This class is used to represent a graph node which consists of multiple jit traced nodes. In a pytorch trace graph,
    there are multiple nodes are traced for one torch.nn.Module object, we group them together to form a single node to
    represent the torch.nn.Module object. We also group some functional call trace nodes together to form a new node.
    """

    def __init__(
        self,
        name,
        unique_name,
        node_type,
        op_type,
        node_cpps,
        inputs=None,
        outputs=None,
        key_node=None,
    ):
        """
        Parameters:
        -----------
        name: str
            node name, such as `conv1`, `backbone.classifier`
        unique_name: str
            A global unique name for current node. Due to some modules,
            such as relu, may be reused several times, so the scopename
            is not suitable as the global unique identifier, so we add a
            unique_name for each node as the global unique identifier.
            We should use the unique_name to traverset the module graph.
        node_type: str
            `module` or `func`
        op_type: str
            operation type, such as `Conv2d`, `aten::view`
        node_cpps: list of torch._C.Node
            jit trace nodes which are included in this new node
        inputs: list of str
            All the inputs of this node, each element is debugName of one input
        outputs: list of str
            All the outputs of this node, each element is debugName of one output
        key_node: torch._C.Node
            The key node of this NodePyGroup.
        """
        super(NodePyGroup, self).__init__(name, [])
        self.node_cpps = node_cpps
        self.name = name
        self.unique_name = unique_name
        self.op_type = op_type
        self.type = node_type
        self.nodes = []
        self.auxiliary = None
        self.add_nodes(node_cpps)
        self.inputs = inputs
        self.outputs = outputs
        # The core node in this NodePyGroup
        self.key_node = key_node

    def add_nodes(self, node_cpps):
        for node_cpp in node_cpps:
            nodepy = NodePyOP(node_cpp)
            nodepy.name = node_cpp.scopeName() + "_" + node_cpp.kind()
            self.nodes.append(nodepy)

    def sub_node_names(self):
        return [x.name for x in self.nodes]

    def __repr__(self):
        return "name: {}, type: {}, op_type: {}, sub_nodes: {}, inputs: {}, outputs: {}, aux: {}".format(
            self.name,
            self.type,
            self.op_type,
            self.sub_node_names(),
            self.inputs,
            self.outputs,
            self.auxiliary,
        )


class TorchModuleGraph(TorchGraph):
    """
    Generates model graph, each node is created from single or multiple jit trace nodes.
    """

    def __init__(self, model=None, dummy_input=None, traced_model=None):
        super().__init__(model, dummy_input, traced_model)
        self.global_count = 0
        self.reused_module = set()
        self.name_to_node, self.input_to_node, self.output_to_node = self._build_graph()
        self._extract_auxiliary_info()

    def _expand_key_func_node(
        self, node, nodes, input_to_node, output_to_node, module_type
    ):
        """
        For trace graph nodes, some nodes are not in modules, these nodes are usually generated by
        the functions directly called in module ```forward```. For such nodes, some of them are
        trivial op which are label by ```prim::```, some of them are not such ops which is call
        non-prim ops. This function is to merge neighbor prim ops to a non-prim op, to construct
        a node.

        Parameters
        ----------
        node : trace graph node
            The non-prim node to expand
        nodes : list of trace graph node
            All the trace graph nodes within the same scope as the non-prim node
        input_to_node : dict
            key: input name, value: a node that uses this input
        output_to_node : dict
            key: output name, value: a node that generates this output
        module_type : str
            can be 'module' or 'func'

        Returns
        -------
        node
            the expanded non-prim node
        """
        # TODO: scope name could be empty
        node_name = ".".join(
            [
                self._get_module_name(node.scopeName()),
                node.kind(),
                str(self.global_count),
            ]
        )
        unique_name = node_name
        _logger.debug("expand non-prim node, node name: %s", node_name)
        self.global_count += 1
        op_type = node.kind()
        node_group = [node]
        inputs = []
        outputs = []
        node_queue = queue.Queue()
        node_queue.put(node)
        while not node_queue.empty():
            curr_node = node_queue.get()
            for _input in curr_node.inputs():
                if _input.node().kind() == CONSTANT_KIND:
                    continue
                input_name = _input.debugName()
                if input_name in output_to_node:
                    for predecessor_node in output_to_node[input_name]:
                        if predecessor_node in nodes:
                            if not self._is_key_func(predecessor_node):
                                if predecessor_node not in node_group:
                                    node_group.append(predecessor_node)
                                    node_queue.put(predecessor_node)
                            else:
                                inputs.append(input_name)
                        else:
                            inputs.append(input_name)
                else:
                    inputs.append(input_name)
        for output in node.outputs():
            if output.node().kind() == CONSTANT_KIND:
                continue
            outputs.append(output.debugName())
        nodepy = NodePyGroup(
            node_name,
            unique_name,
            module_type,
            op_type,
            node_group,
            inputs=inputs,
            outputs=outputs,
            key_node=node,
        )
        return nodepy

    def _expand_module_node(
        self,
        node,
        node_name,
        unique_name,
        op_type,
        nodes,
        input_to_node,
        output_to_node,
        module_type,
    ):
        """
        merge the adjacent nodes of the module. The difference between the
        _expand_module_node and _expand_non_prim_node is that, the _expand_non_prim_node
        only merge the prim:: nodes into the aten:: node, in contrast,the _expand_module_node
        will merge all adjacent nodes into a same nodepy group.

        Parameters
        ----------
        node : trace graph node
            The non-prim node to expand
        node_name : str
            specify the node_name for NodePyGroup
        unique_name : str
            unique_name for the NodePyGroup
        op_type : str
            specify the op_type for the NodePyGroup
        nodes : list of trace graph node
            All the trace graph nodes within the same scope as the non-prim node
        input_to_node : dict
            key: input name, value: a node that uses this input
        output_to_node : dict
            key: output name, value: a node that generates this output
        module_type : str
            can be 'module' or 'func'
        Returns
        -------
        node
            the expanded non-prim node

        """
        _logger.debug("expand module node, node name: %s", node_name)
        self.global_count += 1
        if not op_type:
            op_type = node.kind()
        node_group = [node]
        inputs = []
        outputs = []
        node_queue = queue.Queue()
        node_queue.put(node)
        visited = {node}
        while not node_queue.empty():
            curr_node = node_queue.get()
            for _input in curr_node.inputs():
                if _input.node().kind() == CONSTANT_KIND:
                    continue
                input_name = _input.debugName()
                if input_name in output_to_node:
                    for predecessor_node in output_to_node[input_name]:
                        if predecessor_node in nodes:
                            if predecessor_node not in visited:
                                node_group.append(predecessor_node)
                                node_queue.put(predecessor_node)
                                visited.add(predecessor_node)
                        else:
                            inputs.append(input_name)
                else:
                    inputs.append(input_name)
            for _output in curr_node.outputs():
                if _output.node().kind() == CONSTANT_KIND:
                    continue
                output_name = _output.debugName()
                if output_name in input_to_node:
                    for successor_node in input_to_node[output_name]:
                        if successor_node in nodes:
                            if successor_node not in visited:
                                node_group.append(successor_node)
                                node_queue.put(successor_node)
                                visited.add(successor_node)
                        else:
                            outputs.append(output_name)
                else:
                    outputs.append(output_name)
        unique_outputs = list(set(outputs))
        # remove the dumplicated output names
        unique_outputs.sort(key=outputs.index)

        nodepy = NodePyGroup(
            node_name,
            unique_name,
            module_type,
            op_type,
            node_group,
            inputs=list(inputs),
            outputs=unique_outputs,
        )
        return nodepy

    def _extract_cat_info(self, node_group, cpp_node):
        """
        Extract the detail information of the cat operation,
        such the order of the input tensor, the shape of each
        input tensor, the output shape, and the cat dimension.

        Parameters
        ----------
        node_group : NodePyGroup
        cpp_node: torch._C.Node
            It should be ```aten::cat``` node

        Returns
        -------
        dict
            Include auxiliary information for the cat operation.
            This dict objec has four keys: 'cat_dim', 'out_shape',
            'in_order' and 'in_shape'. cat_dim is the dimension of
            the cat operation to concat the input tensors. out_shape
            is the shape of the output tensor of the cat operation.
            in_order is an ordered list which contains the corresponding
            parent operaion nodes of the input tensors. in_shape is also
            an ordered list that contains the input shapes of the input
            tensor.
        """
        # only suport the cat operation
        assert cpp_node.kind() == CAT_KIND
        cat_info = {}
        # get the shape of the output tensor
        t_output = cpp_node.output()
        out_shape = t_output.type().sizes()
        cat_info["out_shape"] = out_shape
        # get the cat dimension
        inputs = cpp_node.inputs()
        cat_dim = list(inputs)[1].toIValue()
        cat_info["cat_dim"] = cat_dim
        # get the order of the input tensors
        # To get the order of the input tensors, we need
        # to be aware of the topology of the model, which
        # means we should extract the auxiliary information
        # after the build_index function.
        input_order = []
        list_construct_cpp = list(cpp_node.inputs())[0].node()
        input_tensors = list(list_construct_cpp.inputs())
        for _tensor in input_tensors:
            debug_name = _tensor.debugName()
            if debug_name in self.output_to_node:
                input_order.append(self.output_to_node[debug_name].unique_name)
            else:
                # the input tensor may be the input tensor of the whole model
                input_order.append(None)
        cat_info["in_order"] = input_order
        input_shapes = [t.type().sizes() for t in input_tensors]
        cat_info["in_shape"] = input_shapes
        return cat_info

    def _extract_linear_shape_info(self, node_group):
        """
        Extract linear shape input/output tensor shape info from its aten::addmm op.

        Parameters
        ----------
        node_group : NodePyGroup
            NodePyGroup object associated with the linear module.

        Returns
        -------
        dict
            Include shape of input tensor and shape of output tensor
        """
        for cpp_node in node_group.node_cpps:
            if cpp_node.kind() == "aten::addmm":
                # https://github.com/pytorch/pytorch/blob/1.6/torch/nn/functional.py#L1682
                # inputs of aten::addmm:
                # inputs[0] is bias
                # inputs[1] is input data
                # inputs[2] is weight
                t_input = list(cpp_node.inputs())[1]
                t_output = cpp_node.output()
                assert isinstance(t_input.type(), torch._C.TensorType)
                assert isinstance(t_output.type(), torch._C.TensorType)
                in_shape = t_input.type().sizes()
                out_shape = t_output.type().sizes()
                return {"in_shape": in_shape, "out_shape": out_shape}
        return None

    def _extract_shape_info(self, node):
        """
        Extract the shape information of ```aten::view``` node

        Parameters
        ----------
        node : trace graph node
            It should be ```aten::view``` node

        Returns
        -------
        dict
            Include shape of input tensor and shape of output tensor
        """
        t_input = None
        for _input in node.inputs():
            t_input = _input
            break
        t_output = node.output()
        assert isinstance(t_input.type(), torch._C.TensorType)
        assert isinstance(t_output.type(), torch._C.TensorType)
        in_shape = t_input.type().sizes()
        out_shape = t_output.type().sizes()
        return {"in_shape": in_shape, "out_shape": out_shape}

    def _extract_leaf_modules(self):
        """
        Extract leaf modules from the given graph. Leaf module means it does not have submodules.
        To extract leaf modules because only leaf module can be replaced. And shape inference can
        be done in leaf module level. Other shape inference is done in lower level i.e.,
        operation level.

        Returns
        -------
        list
            a list of scope name of all the leaf modules
        """

        def is_parent(name1, name2):
            """
            check if name1 is parent node of name2, for example:
            name1: aa.bb,  name2: aa.bb.cc,  return True
            name1: aa.b,  name2: aa.bb, return False
            """
            parts1, parts2 = name1.split("."), name2.split(".")
            if len(parts1) >= len(parts2):
                return False
            for i, _ in enumerate(parts1):
                if parts2[i] != parts1[i]:
                    return False
            return True

        module_names = sorted([x[0] for x in self.trace.named_modules() if x[0]])
        leaf_nodes = []
        for i, name in enumerate(module_names):
            if i + 1 >= len(module_names) or not is_parent(name, module_names[i + 1]):
                leaf_nodes.append(name)
        return leaf_nodes

    def _get_module_name(self, scope_name):
        """
        Retrieve module name from scope name.
        Parameters:
        -----------
        scope_name: str
            scope_name of a graph node, for example:
            for pytorch 1.3.1: MyModel/BackboneModel[backbone]/Conv2d[conv2]
            for pytorch 1.4.0: __module.backbone/__module.backbone.conv2

        Returns:
        -------
        str
            module name, such as backbone.conv2
        """
        if torch.__version__ >= "1.4.0":
            return scope_name.split("/")[-1].replace("__module.", "")
        else:
            return ".".join(re.findall(r"\[(.*?)\]", scope_name))

    def _build_index(self, nodes_op):
        name_to_node = dict()
        input_to_node = defaultdict(list)
        output_to_node = dict()
        for node in nodes_op:
            name_to_node[node.unique_name] = node
            for _input in node.inputs:
                # inputs may have duplicate tensors
                if node not in input_to_node[_input]:
                    input_to_node[_input].append(node)
            for output in node.outputs:
                if output in output_to_node:
                    assert output_to_node[output] == node, (
                        "One output cannot be generated by multiple nodes %s" % output
                    )
                output_to_node[output] = node
        return name_to_node, input_to_node, output_to_node

    def _is_key_func(self, node_cpp):
        """
        Judge if a cpp node is a key function node.
        If so, we should not merge this node into the
        adjacent node.
        """
        if node_cpp.kind().startswith("aten::"):
            # the nodes that start with 'aten' are key function
            # nodes
            return True
        if node_cpp.kind() in [LIST_UNPACK_KIND, TUPLE_UNPACK_KIND]:
            # We cannot merge the List/Tuple
            # Unpack func into other nodes, else it
            # may lead to a graph construction error.
            # The reason why we donnot take the construct node
            # also as a key node is that `cat` operation node need
            # the last(previous) visited node to infer the mask. If
            # we take the Construct node as the important node, the
            # predecessor of the `cat` node will always be a construct
            # node, which means we cannot infer the mask for the cat
            # operation.
            return True
        return False

    def unpack_manually(self):
        """
        Unpack the tensor tuple or tensor list manually,
        and remove the ListUnpack/TupleUnpack node from
        the graph. Note: this function will change the
        graph structure.
        """
        if hasattr(self, "unpacked"):
            # if already unpacked the tuple/list manually
            return
        for node in self.nodes_py.nodes_op:
            if node.op_type in [TUPLE_UNPACK_KIND, LIST_UNPACK_KIND]:
                unpack_cpp = node.key_node
                last_cpp = list(unpack_cpp.inputs())[0].node()
                if last_cpp.kind() in [TUPLE_CONSTRUCT_KIND, LIST_CONSTRUCT_KIND]:
                    # we need check if the tensor tuple or tensor list is produced
                    # by a list/tuple construct node. If so, we can unpack the tuple
                    # or list manunally.
                    _logger.debug("List/Tuple Construct Node(cpp) %s", str(last_cpp))
                    _logger.debug("List/Tuple Unpack Node(cpp) %s", str(unpack_cpp))
                    assert len(list(unpack_cpp.outputs())) == len(
                        list(last_cpp.inputs())
                    )
                    errmsg = (
                        "%s Input number: %d if inconsistent with the output number %d"
                        % (unpack_cpp, len(node.inputs), len(list(last_cpp.inputs())))
                    )

                    assert len(node.inputs) == len(list(last_cpp.inputs())), errmsg
                    for _debug_input, _debug_output in zip(node.inputs, node.outputs):
                        if (
                            _debug_input in self.input_to_node
                            and _debug_output in self.input_to_node
                        ):
                            # input_to_node[_debug_input] is a list of NodePyGroup, because
                            # one tensor can be used as input for multiple nodes at the same time.

                            # note that, in this case, the construct cpp node and unpack cpp node
                            # will be merged into the same NodePyGroup, so we remove the `node` from
                            # input_to_node[_debug_input] and directly connect this tensor to the
                            # input_to_node[_debug_output]
                            if node in self.input_to_node[_debug_input]:
                                self.input_to_node[_debug_input].remove(node)
                            # add the following nodes of _output into the input_to_node[_debug_input]
                            self.input_to_node[_debug_input].extend(
                                self.input_to_node[_debug_output]
                            )
                        # just remove the _debug_output from the grapgh index. So that we can also skip
                        # the construct and tuple
                        if _debug_output in self.input_to_node:
                            for following_node in self.input_to_node[_debug_output]:
                                _tmp_index = following_node.inputs.index(_debug_output)
                                following_node.inputs[_tmp_index] = _debug_input

        self.unpacked = True

    def _build_graph(self):
        """
        Build graph using our defined format from jit trace.
        There are basically three steps: first, construct necessary information (data structures),
        second, extract all the modules to convert to node, Third, extract all functions to convert
        to node.

        Returns
        -------
        dict
            use name to index nodes, key: node name, value: node
        dict
            use input (its name) to index nodes,
            key: input, value: list of nodes that take this input
        dict
            use output (its name) to index nodes,
            key: output, value: node that generates this output
        """
        omit_useless_nodes = True
        graph = self.trace.graph
        _logger.debug(graph)
        # build input/output mapping, from input/output debugName to its node
        input_to_node = defaultdict(list)
        output_to_node = defaultdict(list)
        for node in graph.nodes():
            if node.kind() == CONSTANT_KIND:
                continue
            for x in node.outputs():
                if x.node().kind() == CONSTANT_KIND:
                    continue
                output_to_node[x.debugName()].append(node)
                assert len(output_to_node[x.debugName()]) <= 1, (
                    "One output cannot be generated by multiple nodes %s"
                    % x.debugName()
                )
            for x in node.inputs():
                if x.node().kind() == CONSTANT_KIND:
                    continue
                input_to_node[x.debugName()].append(node)

        # build module mapping, from module name to all nodes (as list) under this module scope
        module_to_nodes = defaultdict(list)
        # the mapping of function (non-module in forward) to nodes, key is scope name
        func_to_nodes = defaultdict(list)

        nodes_py = GraphPy()
        for node in graph.inputs():
            if omit_useless_nodes:
                if (
                    not node.uses()
                ):  # number of user of the node (= number of outputs/ fanout)
                    continue

            if node.type().kind() != "ClassType":
                nodes_py.append(NodePyIO(node, "input"))

        self.leaf_modules = self._extract_leaf_modules()
        module_to_type = {
            name: parse_traced_name(
                module._name if hasattr(module, "_name") else module.original_name
            )
            for name, module in self.trace.named_modules()
        }

        # associate module name with their trace graph nodes
        for node in graph.nodes():
            if node.kind() == CONSTANT_KIND:
                continue
            module_name = self._get_module_name(node.scopeName())
            if module_name in self.leaf_modules:
                module_to_nodes[module_name].append(node)
            else:
                func_to_nodes[node.scopeName()].append(node)
        # build node group for module
        for module_name, node_cpps in module_to_nodes.items():
            use_count = 0
            merged = set()
            for node in node_cpps:
                if node not in merged:
                    # modules that have same scope name may have different locations in the
                    # graph. Futhermore, there are also lots of prim:: nodes that in node_cpps,
                    # so we also need to call the expand_module_node.
                    unique_name = module_name
                    if use_count > 0:
                        unique_name = module_name + ".%d" % use_count
                        self.reused_module.add(unique_name)
                        self.reused_module.add(module_name)
                    node_group = self._expand_module_node(
                        node,
                        module_name,
                        unique_name,
                        module_to_type[module_name],
                        node_cpps,
                        input_to_node,
                        output_to_node,
                        "module",
                    )
                    nodes_py.nodes_op.append(node_group)
                    use_count += 1
                    merged.update(node_group.node_cpps)

        # each scope_name may have multiple funcs, we split them and create node for each of them
        # build node group for torch.nn.functional
        for _, nodes in func_to_nodes.items():
            # extract non prim:: nodes
            key_func_nodes = list()
            for node in nodes:
                if self._is_key_func(node):
                    # find the key function nodes
                    key_func_nodes.append(node)
            # for each non prim node, expand it
            for node in key_func_nodes:
                node_group = self._expand_key_func_node(
                    node, nodes, input_to_node, output_to_node, "func"
                )
                nodes_py.nodes_op.append(node_group)
                # get shape infor for view (aten::view) func
                # if node_group.op_type in ['aten::view', 'aten::flatten']:
                #     node_group.auxiliary = self._extract_shape_info(node)

        for node in graph.outputs():  # Create sink nodes for output ops
            node_py = NodePyIO(node, "output")
            nodes_py.append(node_py)

        self.nodes_py = nodes_py
        # build index
        return self._build_index(self.nodes_py.nodes_op)

    def _extract_auxiliary_info(self):
        """
        Extract the auxiliary information for the nodegroups
        if necessary. For example, view/flatten operations may
        need the shape of the input tensor and output tensor.
        """
        # extract the input & output shape for the view and flatten
        for node_group in self.nodes_py.nodes_op:
            if node_group.op_type in [
                "aten::view",
                "aten::flatten",
                "aten::mean",
                "aten::reshape",
                "aten::expand_as",
                "aten::pixel_shuffle",
            ]:
                # get shape infor for view (aten::view) func
                cpp_node = list(
                    filter(
                        lambda x: x.kind() == node_group.op_type, node_group.node_cpps
                    )
                )[0]
                node_group.auxiliary = self._extract_shape_info(cpp_node)
            elif node_group.op_type == "Linear":
                node_group.auxiliary = self._extract_linear_shape_info(node_group)
            elif node_group.op_type == CAT_KIND:
                # get the detail information for cat func
                cpp_node = list(
                    filter(
                        lambda x: x.kind() == node_group.op_type, node_group.node_cpps
                    )
                )[0]
                node_group.auxiliary = self._extract_cat_info(node_group, cpp_node)

    def find_predecessors(self, unique_name):
        """
        Find predecessor node of the given node

        Parameters
        ----------
        unique_name : str
            The unique name of the node

        Returns
        -------
        list
            a list of nodes who are the given node's predecessor
        """
        predecessors = []
        for _input in self.name_to_node[unique_name].inputs:
            if not _input in self.output_to_node:
                _logger.debug("cannot find node with %s as its output", _input)
            else:
                node_py = self.output_to_node[_input]
                predecessors.append(node_py.unique_name)
        return predecessors

    def find_successors(self, unique_name):
        """
        Find successor nodes of the given node

        Parameters
        ----------
        unique_name : str
            The unique name of the node

        Returns
        -------
        list
            a list of nodes who are the given node's successor
        """
        successors = []
        for output in self.name_to_node[unique_name].outputs:
            if output not in self.input_to_node:
                # may reach the output of the whole graph
                continue
            nodes_py = self.input_to_node[output]
            for node_py in nodes_py:
                successors.append(node_py.unique_name)
        return successors



"""try:
    import torch
except ModuleNotFoundError:
    raise ModuleNotFoundError(
        "PyTorch not installed. "
        "Please appropriately install PyTorch, "
        "see https://pytorch.org/ for installation."
    )"""

from setuptools import setup, find_packages

with open("README.md", 'r') as fh:
    long_description = fh.read()

''' https://packaging.python.org/guides/distributing-packages-using-setuptools/ '''
''' https://setuptools.readthedocs.io/en/latest/ '''
setup(
    name='autogllight',
    version='0.1.1',
    author='THUMNLab/aglteam',
    maintainer='THUMNLab/aglteam',
    author_email='autogl@tsinghua.edu.cn',
    description='AutoML tools for graph-structure dataset',
    long_description=long_description,
    long_description_content_type='text/markdown',
    include_package_data=True,
    packages=find_packages(),
    # https://packaging.python.org/guides/distributing-packages-using-setuptools/#python-requires
    python_requires='>=3.6',
    # https://pypi.org/classifiers/
    classifiers=[
        "Development Status :: 4 - Beta",
        "License :: OSI Approved :: Apache Software License",
        "Programming Language :: Python :: 3 :: Only",
        "Programming Language :: Python :: 3.6"
    ],
    # https://setuptools.readthedocs.io/en/latest/userguide/dependency_management.html
    # note that setup_requires and tests_require are deprecated
    install_requires=[
        'pyyaml',
        'pytest',
    ]
)


"""
HPO Module for tuning hyper parameters
"""

import random
import time

import numpy as np
# from autogl.backend import DependentBackend
from tqdm import trange

# from ..feature import NetLSD as SgNetLSD
from . import register_hpo
from .autone_file import utils
from .base import BaseHPOptimizer, TimeTooLimitedError

# _isdgl = DependentBackend.is_dgl()
_isdgl = False
if _isdgl:
    import dgl
else:
    from torch_geometric.data import GraphSAINTRandomWalkSampler, InMemoryDataset

    class _MyDataset(InMemoryDataset):
        def __init__(self, datalist) -> None:
            super().__init__()
            self.data, self.slices = self.collate(datalist)


@register_hpo("autone")
class AutoNE(BaseHPOptimizer):
    """
    AutoNE HPOptimizer
    The Implementation of "AutoNE: Hyperparameter Optimization for Massive Network Embedding"(KDD 2019).
    See https://github.com/tadpole/AutoNE for more information

    Attributes
    ----------
    max_evals : int
        The max rounds of evaluating HPs
    subgraphs : int
        The number of subgraphs
    sub_evals : int
        The number of evaluation times on each subgraph
    sample_batch_size, sample_walk_length : int
        Using for sampling subgraph, see torch_geometric.data.GraphSAINRandomWalkSampler
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.max_evals = kwargs.get("max_evals", 100)
        self.subgraphs = kwargs.get("subgraphs", 2)
        self.sub_evals = kwargs.get("sub_evals", 2)
        self.sample_batch_size = kwargs.get("sample_batch_size", 150)
        self.sample_walk_length = kwargs.get("sample_walk_length", 100)
        self.dataset = kwargs.get("dataset", None)

    def optimize(self):
        """
        Optimize the HP by the method within give model and HP space

        See .base.BaseHPOptimizer.optimize
        """
        if self.dataset is None:
            raise ValueError("dataset is None")

        dataset = self.dataset

        space = self.hp_space
        current_space = self._encode_para(space)

        def sample_subgraph(whole_data):
            data = whole_data.data
            loader = GraphSAINTRandomWalkSampler(
                data,
                batch_size=self.sample_batch_size,
                walk_length=self.sample_walk_length,
                num_steps=self.subgraphs,
                save_dir=whole_data.processed_dir,
            )
            results = []
            for data in loader:
                in_dataset = _MyDataset([data])
                results.append(in_dataset)
            return results

        def sample_subgraph_dgl(whole_data):
            data = whole_data[0]  # dgl data
            # find data with different labels
            # random walk
            start = [
                random.randint(0, data.num_nodes - 1) for i in range(self.subgraphs)
            ]
            traces, _ = dgl.sampling.random_walk_with_restart(
                data,
                start,
                length=self.sample_batch_size,
                restart_prob=1 / self.sample_walk_length,
            )
            subgraphs = dgl.node_subgraph(data, [traces[i, :] for i in traces.size(0)])
            return subgraphs

        func = SgNetLSD()

        def get_wne(graph):
            graph = func.fit_transform(graph)
            # transform = nx.NxGraph.compose(map(lambda x: x(), nx.NX_EXTRACTORS))
            # print(type(graph))
            # gf = transform.fit_transform(graph).data.gf
            gf = graph.data.gf
            fin = list(gf[0]) + list(map(lambda x: float(x), gf[1:]))
            return fin

        start_time = time.time()

        # code in AutoNE
        sampled_number = self.subgraphs
        k = self.sub_evals
        s = self.max_evals
        X = []
        y = []
        params = utils.Params()
        params.set_space(current_space)
        total_t = 0.0
        info = []
        K = utils.K(len(params.type_))
        gp = utils.GaussianProcessRegressor(K)
        if _isdgl:
            sample_graphs = sample_subgraph_dgl(dataset)
        else:
            sample_graphs = sample_subgraph(dataset)
        print("Sample Phase:\n")
        for t in trange(sampled_number):
            b_t = time.time()
            i = t
            subgraph = sample_graphs[t]
            wne = get_wne(subgraph)
            for v in range(k):
                kargs = params.random_x()
                para = params.x2dict(kargs)
                externel_para, trial_para = self._decode_para(para)
                perf = self.f(externel_para, subgraph)
                X_reg = params.dict2x(trial_para)
                X.append(np.hstack((X_reg, wne)))
                y.append(perf)

        best_perf = None
        best_hp = None
        wne = get_wne(dataset)
        print("HPO Search Phase:\n")
        for t in trange(s):
            if time.time() - start_time > self.time_limit:
                self.logger.info("Time out of limit, Epoch: {}".format(str(i)))
                break
            b_t = time.time()
            gp.fit(np.vstack(X), y)
            X_temp, _ = gp.predict(params.get_bound(), params.get_type(), wne)
            X_temp = X_temp[: len(params.type_)]
            para = params.x2dict(X_temp)
            externel_para, trial_para = self._decode_para(para)
            perf_temp = self.f(externel_para, dataset)
            self.trials.append(
                self._creat_a_trail(
                    "HPO", "Autone", "Completed", externel_para, perf_temp
                )
            )
            # self._print_info(externel_para, perf_temp)
            X_reg = params.dict2x(trial_para)

            X.append(np.hstack((X_reg, wne)))
            y.append(perf_temp)
            if not best_perf or perf_temp < best_perf:
                best_perf = perf_temp
                best_hp = externel_para

            e_t = time.time()
            total_t += e_t - b_t

        if not best_perf:
            raise TimeTooLimitedError(
                "Given time is too limited to finish one round in HPO."
            )

        # self.logger.info("Best Parameter:")
        # self._print_info(best_hp, best_perf)

        return best_hp, best_perf



from ...suggestion.algorithm.base_skopt_algorithm import BaseSkoptAlgorithm


class SkoptBayesianOptimization(BaseSkoptAlgorithm):
    """
    The implementation is based on https://github.com/tobegit3hub/advisor
    Get the new suggested trials with bayesian optimization algorithm.
    """

    def __init__(self):
        super(SkoptBayesianOptimization, self).__init__("bayesian_optimization")



import torch
import torch.nn.functional as F
from autogllight.utils.evaluation import Acc

from ..space import BaseSpace
from .base import BaseEstimator


class OneShotOGBEstimator(BaseEstimator):
    """
    One shot estimator on ogb data

    Use model directly to get estimations.

    Parameters
    ----------
    loss_f : str
        The name of loss funciton in PyTorch
    evaluation : list of Evaluation
        The evaluation metrics in module/train/evaluation
    """

    def __init__(self, loss_f="nll_loss", evaluation=[Acc()]):
        super().__init__(loss_f, evaluation)
        self.evaluation = evaluation

    def infer(self, model: BaseSpace, dataloader, *args, **kwargs):
        device = next(model.parameters()).device
        y_true = []
        y_pred = []
        for batch in dataloader:
            batch = batch.to(device)
            _, pred, _, _ = model(batch)
            y_true.append(batch.y.view(pred.shape).detach().cpu())
            y_pred.append(pred.detach().cpu())

        y_true = torch.cat(y_true, dim=0).float()
        y_pred = torch.cat(y_pred, dim=0)

        loss = getattr(F, self.loss_f)(y_pred, y_true).item()

        y_true = y_true.view(-1).numpy()
        y_pred = y_pred.view(-1).numpy()

        metrics = {
            eva.get_eval_name(): eva.evaluate(y_pred, y_true) for eva in self.evaluation
        }
        return metrics, loss



from suggestion.early_stop_algorithm.abstract_early_stop import (
    AbstractEarlyStopAlgorithm,
)


class EarlyStopFirstTrialAlgorithm(AbstractEarlyStopAlgorithm):
    def get_early_stop_trials(self, trials):
        return [trials[0]]



import torch
import torch.nn as nn
import torch.nn.functional as F
import numpy as np
from torch_geometric.utils import add_self_loops,remove_self_loops
# from operations import *
from .op_graph_classification import *
from torch.autograd import Variable
from .genotypes import NA_PRIMITIVES, LA_PRIMITIVES, POOL_PRIMITIVES, READOUT_PRIMITIVES, ACT_PRIMITIVES
# from genotypes import Genotype
from torch_geometric.nn import  global_mean_pool,global_add_pool
from ogb.graphproppred.mol_encoder import AtomEncoder,BondEncoder

class NaSingleOp(nn.Module):
  def __init__(self, in_dim, out_dim, with_linear):
    super().__init__()
    self._ops = nn.ModuleList()

    self.op = NA_OPS['gin'](in_dim, out_dim)

    self.op_linear = torch.nn.Linear(in_dim, out_dim)

  def forward(self, x, edge_index, edge_weights, edge_attr, with_linear):
    mixed_res = []
    if with_linear:
        mixed_res.append(self.op(x, edge_index, edge_weight=edge_weights, edge_attr=edge_attr)+self.op_linear(x))
        # print('with linear')
    else:
        mixed_res.append(self.op(x, edge_index, edge_weight=edge_weights, edge_attr=edge_attr))
        # print('without linear')
    return sum(mixed_res)

class NaDisenOp(nn.Module):
  def __init__(self, in_dim, out_dim, with_linear, k = 4):
    super().__init__()
    self.ops = nn.ModuleList()
    self.op_linear = nn.ModuleList()
    self.in_dim = in_dim
    self.k = k

    for i in range(k):
      self.ops.append(NA_OPS['gin'](in_dim // 4, out_dim // 4))
      self.op_linear.append(torch.nn.Linear(in_dim, out_dim))

  def forward(self, x, edge_index, edge_weights, edge_attr, with_linear):
    # x: node * d
    mixed_res = []
    xs = x.hsplit(self.k)
    for i in range(self.k):
      z = self.ops[i](xs[i], edge_index, edge_weight=edge_weights, edge_attr=edge_attr)
      if with_linear:
        z = z + self.op_linear[i](xs[i])
      mixed_res.append(z)
    res = torch.hstack(mixed_res)
    return res

class Disen3Head(nn.Module):
  def __init__(self, in_dim, k = 4):
    super().__init__()
    self.ops = nn.ModuleList()
    self.in_dim = in_dim
    self.k = k

    for i in range(3):
      self.ops.append(torch.nn.Linear(in_dim // 4, 1))

  def forward(self, x):
    # x: node * d
    mixed_res = []
    xs = x.hsplit(self.k)
    for i in range(3):
      z = self.ops[i](xs[i])
      z = 0.05 + 0.35 * torch.sigmoid(z)
      mixed_res.append(z)
    res = torch.hstack(mixed_res)
    return res

class GEncoder(nn.Module):
  def __init__(self, criterion, in_dim, out_dim, hidden_size, num_layers=2, dropout=0.5, epsilon=0.0, args=None, with_conv_linear=False,num_nodes=0, mol = False, virtual = False):
    super().__init__()

    self.in_dim = in_dim
    self.out_dim = out_dim
    self.hidden_size = hidden_size
    self.num_layers = num_layers
    self.num_nodes = num_nodes
    self._criterion = criterion
    self.dropout = dropout
    self.epsilon = epsilon
    self.with_linear = with_conv_linear
    self.explore_num = 0
    self.args = args
    self.temp = args.temp
    self._loc_mean = args.loc_mean
    self._loc_std = args.loc_std
    self.mol = mol # if the task is molecule
    self.virtual = virtual

    self.lin1 = nn.Linear(in_dim, hidden_size)
    self.atom_encoder = AtomEncoder(hidden_size)
    self.virtualnode_embedding = torch.nn.Embedding(1, hidden_size)
    torch.nn.init.constant_(self.virtualnode_embedding.weight.data, 0)

    self.mlp_virtualnode_list = torch.nn.ModuleList()
    for layer in range(num_layers - 1):
        self.mlp_virtualnode_list.append(torch.nn.Sequential(torch.nn.Linear(hidden_size, 2*hidden_size), torch.nn.BatchNorm1d(2*hidden_size), torch.nn.ReLU(), \
                                                torch.nn.Linear(2*hidden_size, hidden_size), torch.nn.BatchNorm1d(hidden_size), torch.nn.ReLU()))

    # node aggregator op
    self.gnn_layers = nn.ModuleList()
    for i in range(num_layers):
        if i < 1:
          self.gnn_layers.append(NaSingleOp(hidden_size, hidden_size, self.with_linear))
        else:
          self.gnn_layers.append(NaDisenOp(hidden_size, hidden_size, self.with_linear))

    #self.pooling_global = PoolingMixedOp(hidden_size * (num_layers + 1), args.pooling_ratio, num_nodes=self.num_nodes)
    self.pooling_trivial = Pooling_trivial(hidden_size * (num_layers + 1), args.pooling_ratio)

    #graph representation aggregator op
    #self.layer6 = LaMixedOp(hidden_size, num_layers+1)
    self.layer7 = Readout_trivial()
    self.disenhead = Disen3Head(hidden_size)

    self.lin_output = nn.Linear(hidden_size * 2 * (num_layers + 1), hidden_size)
    self.classifier = nn.Linear(hidden_size, out_dim)

  def forward(self, data, discrete=False, mode='none'):
    self.args.search_act = False
    with_linear = self.with_linear
    x, edge_index = data.x, data.edge_index
    edge_attr = getattr(data, 'edge_attr', None)
    batch = data.batch
    # edge_index, _ = remove_self_loops(edge_index)
    if edge_attr == None:
        edge_index, _ = add_self_loops(edge_index, num_nodes=x.size()[0])

    try:
        x = F.elu(self.lin1(x))
    except RuntimeError:
        x = self.atom_encoder(x)

    edge_weights = torch.ones(edge_index.size()[1], device=edge_index.device).float()
    virtualnode_embedding = self.virtualnode_embedding(torch.zeros(batch[-1].item() + 1).to(edge_index.dtype).to(edge_index.device))

    #graph_representations.append(self.readout_layers[0](x, batch, None, readout_alphas[0]))

    #x = F.dropout(x, p=self.dropout, training=self.training)
    gr = [x]

    for i in range(self.num_layers):
        if self.virtual:
            orix = x
            x = x + virtualnode_embedding[batch]
        x = self.gnn_layers[i](x, edge_index, edge_weights, edge_attr, with_linear)
        #print('evaluate data {}-th gnn:'.format(i), x.size(), batch.size())

        x = F.elu(x)
        layer_norm = nn.LayerNorm(normalized_shape=x.size(), elementwise_affine=False)
        x = layer_norm(x)
        x = F.dropout(x, p=self.dropout, training=self.training)
        gr.append(x)

        if self.virtual and i < self.num_layers - 1:
            virtualnode_embedding_temp = global_add_pool(orix, batch) + virtualnode_embedding
            ### transform virtual nodes using MLP
            virtualnode_embedding = F.dropout(self.mlp_virtualnode_list[i](virtualnode_embedding_temp), self.dropout, training = self.training)

    gr = torch.cat(gr, 1)

    # if use pooling trivial, you should set the GNN in SAG in op_graph_classification.py
    x, edge_index, edge_weights, batch, _ = self.pooling_trivial(gr, edge_index, edge_weights, data, batch, None)
    #x, edge_index, edge_weights, batch, _ = self.pooling_global(gr, edge_index, edge_weights, data, batch, None, pool_alphas[0])
    #print(batch)

    x5 = self.layer7(x, batch)
    #x5 = self.layer6(graph_representations, la_alphas[0])

    x5 = self.lin_output(x5)
    x5 = F.elu(x5)
    x_emb = x5
    ssloutput = self.disenhead(x_emb)
    #x5 = F.dropout(x5, p=self.dropout, training=self.training)
    logits = self.classifier(x5)

    if not self.mol:
        logits = F.log_softmax(logits, dim = -1)

    return logits, x_emb, ssloutput

  def arch_parameters(self):
    return self._arch_parameters



from . import BaseSpace
from .graces_space import *


class GracesSpace(BaseSpace):
    def __init__(self, input_dim, output_dim, num_nodes, mol, virtual, criterion, args):
        super().__init__()
        self.input_dim = input_dim
        self.output_dim = output_dim
        self.num_nodes = num_nodes
        self.mol = mol
        self.virtual = virtual
        self.criterion = criterion
        self.args = args
        self.use_forward = True

    def build_graph(self):
        self.supernet0 = GEncoder(
            criterion=self.criterion,
            in_dim=self.input_dim,
            out_dim=self.output_dim,
            hidden_size=self.args.graph_dim,
            num_layers=2,
            dropout=0.5,
            epsilon=self.args.epsilon,
            args=self.args,
            with_conv_linear=self.args.with_conv_linear,
            num_nodes=self.num_nodes,
            mol=self.mol,
            virtual=self.virtual,
        )
        self.supernet = Network(
            criterion=self.criterion,
            in_dim=self.input_dim,
            out_dim=self.output_dim,
            hidden_size=self.args.hidden_size,
            num_layers=self.args.num_layers,
            dropout=self.args.dropout,
            epsilon=self.args.epsilon,
            args=self.args,
            with_conv_linear=self.args.with_conv_linear,
            num_nodes=self.num_nodes,
            mol=self.mol,
            virtual=self.virtual,
        )
        num_na_ops = len(NA_PRIMITIVES)
        num_pool_ops = len(POOL_PRIMITIVES)
        self.ag = AG(args=self.args, num_op=num_na_ops, num_pool=num_pool_ops)
        self.explore_num = 0

    def forward(self, data):
        if not self.use_forward:
            return self.prediction
        # mhssl
        pred0, graph_emb, sslout = self.supernet0(data, mode="mixed")
        # ag
        graph_alpha, cosloss = self.ag(graph_emb)
        # final supernet
        pred, _ = self.supernet(data, mode="mads", graph_alpha=graph_alpha)
        self.current_pred = pred
        return pred0, pred, cosloss, sslout

    def keep_prediction(self):
        self.prediction = self.current_pred

    def parse_model(self, selection):
        self.use_forward = False
        return self.wrap()



import torch.nn.functional as F
from torch_scatter import scatter
from torch_sparse import SparseTensor
from torch_geometric.utils import add_remaining_self_loops
from torch_geometric.nn import SAGPooling, TopKPooling,EdgePooling,ASAPooling,dense_diff_pool,LEConv,GINConv,GraphConv,GCNConv
from torch_geometric.nn.pool.topk_pool import topk, filter_adj
from torch_geometric.nn.conv import MessagePassing
import torch
from torch.nn import Linear, Sequential, ReLU,ELU, BatchNorm1d as BN
from torch_scatter import scatter_add, scatter_max
from torch_geometric.utils import softmax
from torch_geometric.nn.inits import reset
from torch_geometric.nn.pool.topk_pool import topk

class ASAPooling_mix(ASAPooling):

    def forward(self, x, edge_index, edge_weight=None, batch=None, add_self_loop=False, remove_self_loop=False, ft=False):
        N = x.size(0)

        if self.add_self_loops:
            edge_index, edge_weight = add_remaining_self_loops(
                edge_index, edge_weight, fill_value=1)
                # edge_index, edge_weight, fill_value=1, num_nodes=N)
        if edge_weight==None:
            edge_weight = torch.ones(edge_index.size()[1], device=x.device)

        if batch is None:
            batch = edge_index.new_zeros(x.size(0))

        x_pool = x
        if self.GNN is not None:
            x_pool = self.gnn_intra_cluster(x=x, edge_index=edge_index,
                                            edge_weight=edge_weight)

        x_pool_j = x_pool[edge_index[0]]
        x_q = scatter(x_pool_j, edge_index[1], dim=0, reduce='max') #Eq.5 in ASAP
        x_q = self.lin(x_q)[edge_index[1]]  #Wm_i in Eq.6

        score = self.att(torch.cat([x_q, x_pool_j], dim=-1)).view(-1)  #W * \sigma(Wm_i || xj) in Eq.6
        score = F.leaky_relu(score, self.negative_slope)
        score = softmax(score, edge_index[1], num_nodes=N) #Eq.6

        # Sample attention coefficients stochastically.
        score = F.dropout(score, p=self.dropout, training=self.training)

        v_j = x[edge_index[0]] * score.view(-1, 1)
        x = scatter(v_j, edge_index[1], dim=0, reduce='add') #Eq.7

        # Cluster selection.
        fitness = self.gnn_score(x, edge_index, edge_weight=edge_weight).sigmoid().view(-1)
        perm = topk(fitness, self.ratio, batch)
        if ft:
            x = x[perm] * fitness[perm].view(-1, 1)
            batch = batch[perm]
            # edge_index, edge_attr = filter_adj(edge_index, None, perm,
            #                                    num_nodes=score.size(0))
            # Graph coarsening.
            row, col = edge_index
            A = SparseTensor(row=row, col=col, value=edge_weight,
                             sparse_sizes=(N, N))
            S = SparseTensor(row=row, col=col, value=score, sparse_sizes=(N, N))
            S = S[:, perm]

            A = S.t() @ A @ S

            if self.add_self_loops:
                A = A.fill_diag(1.)
            else:
                A = A.remove_diag()

            row, col, edge_weight = A.coo()
            edge_index = torch.stack([row, col], dim=0)

            return x, edge_index, edge_weight, batch, perm

        else:
            mask = torch.zeros_like(fitness)
            mask[perm] = 1

            x1 = x * fitness.view(-1, 1) #x:[node_num, feature_dim] mask:[node_num]
            x2 = x1 * mask.view(-1, 1) #for these unselected nodes, set features to zero

            new_edge_weights = mask[edge_index[0]]+mask[edge_index[1]]
            edges_mask = torch.where(new_edge_weights != 2, torch.zeros(1, device=edge_index.device), torch.ones(1, device=edge_index.device))
            edge_weight2 = edge_weight * edges_mask

            return x2, edge_index, edge_weight2, batch, perm

class SAGPool_mix(torch.nn.Module):
    def __init__(self, in_channels, ratio=0.5, gnn_type='gcn', min_score=None,
                 multiplier=1, nonlinearity=torch.tanh, **kwargs):
        super(SAGPool_mix, self).__init__()
        self.gnn_type = gnn_type
        self.in_channels = in_channels
        self.ratio = ratio
        if gnn_type == 'gcn':
            self.gnn = GCNConv(in_channels, 1)
        elif gnn_type == 'mlp':
            self.gnn = Sequential(
                Linear(in_channels, int(in_channels//2)),
                ELU(),
                Linear(int(in_channels//2), 1))
        else:
            self.gnn = GraphConv(in_channels, 1)
        self.min_score = min_score
        self.multiplier = multiplier
        self.nonlinearity = nonlinearity

        self.reset_parameters()

    def reset_parameters(self):
        reset(self.gnn)

    def forward(self, x, edge_index, edge_weight=None, edge_attr=None, batch=None, attn=None, add_self_loop=False, remove_self_loop=False, ft=False):
        """"""
        if add_self_loop:
            edge_index, edge_weight = add_remaining_self_loops(
                edge_index, edge_weight, fill_value=1)

        if edge_weight == None:
            edge_weight = torch.ones(edge_index.size()[1], device=x.device)

        if batch is None:
            batch = edge_index.new_zeros(x.size(0))

        attn = x if attn is None else attn
        attn = attn.unsqueeze(-1) if attn.dim() == 1 else attn
        if self.gnn_type == 'mlp':
            score = self.gnn(attn).view(-1)
        else:
            score = self.gnn(attn, edge_index, edge_weight=edge_weight).view(-1)

        if self.min_score is None:
            score = self.nonlinearity(score)
        else:
            score = softmax(score, batch)

        perm = topk(score, self.ratio, batch, self.min_score)
        if ft:
            x = x[perm] * score[perm].view(-1, 1)
            x = self.multiplier * x if self.multiplier != 1 else x
            batch = batch[perm]
            edge_index, edge_attr = filter_adj(edge_index, edge_attr, perm,
                                               num_nodes=score.size(0))
            return x, edge_index, edge_weight, batch, perm
        else:
            mask = torch.zeros_like(score)
            mask[perm] = 1
            x1 = x * score.view(-1, 1)
            x2 = x1 * mask.view(-1, 1) #for these unselected nodes, set features to zero

            #edge_weights
            new_edge_weights = mask[edge_index[0]] + mask[edge_index[1]]
            edges_mask = torch.where(new_edge_weights != 2, torch.zeros(1, device=edge_index.device), torch.ones(1, device=edge_index.device))
            edge_weight2 = edge_weight * edges_mask


            return x2, edge_index, edge_weight2, batch, perm

class TOPKpooling_mix(TopKPooling):
    def forward(self, x, edge_index, edge_weight=None, edge_attr=None, batch=None, attn=None, add_self_loop=False, remove_self_loop=False, ft=False):
        """"""
        if batch is None:
            batch = edge_index.new_zeros(x.size(0))
        if add_self_loop:
            edge_index, edge_weight = add_remaining_self_loops(
                edge_index, edge_weight, fill_value=1)

        if edge_weight == None:
            edge_weight = torch.ones(edge_index.size()[1], device=x.device)

        attn = x if attn is None else attn
        attn = attn.unsqueeze(-1) if attn.dim() == 1 else attn
        score = (attn * self.weight).sum(dim=-1)

        if self.min_score is None:
            score = self.nonlinearity(score / self.weight.norm(p=2, dim=-1))
        else:
            score = softmax(score, batch)

        perm = topk(score, self.ratio, batch, self.min_score)
        if ft:
            x = x[perm] * score[perm].view(-1, 1)
            x = self.multiplier * x if self.multiplier != 1 else x

            batch = batch[perm]
            edge_index, edge_attr = filter_adj(edge_index, edge_attr, perm,
                                               num_nodes=score.size(0))

            return x, edge_index, edge_weight, batch, perm
        else:
            mask = torch.zeros_like(score)
            mask[perm] = 1
            x1 = x * score.view(-1, 1)
            x2 = x1 * mask.view(-1, 1) #for these unselected nodes, set features to zero

            #edge_weights
            new_edge_weights = mask[edge_index[0]]+mask[edge_index[1]]
            edges_mask = torch.where(new_edge_weights != 2, torch.zeros(1, device=edge_index.device), torch.ones(1, device=edge_index.device))
            edge_weight2 = edge_weight * edges_mask

            return x2, edge_index, edge_weight2, batch, perm

class Hoppooling_mix(torch.nn.Module):
    def __init__(self, in_channels, ratio, walk_length=3):
        super(Hoppooling_mix, self).__init__()
        self.walk_length = walk_length
        self.pooling_ratio = ratio
    def reset_parameters(self):
        pass
    def forward(self, x, edge_index, batch=None, edge_weight=None, edge_attr=None,ft=False):
        if batch is None:
            batch = edge_index.new_zeros(x.size(0))
        if edge_weight == None:
            edge_weight = torch.ones(edge_index.size()[1], device=x.device)
        k_hops=[]
        num_nodes_1hop = scatter_add(edge_weight, edge_index[0], dim=0)
        k_hops.append(num_nodes_1hop)
        for i in range(int(self.walk_length) - 1):
            num_nodes_1hop = scatter_add(num_nodes_1hop[edge_index[1]] * edge_weight, edge_index[0], dim=0)
            k_hops.append(num_nodes_1hop)

        # x = scatter_add(x[edge_index[1]] * edge_weight, edge_index[0], dim=0)
        # score = num_nodes_1hop
        score = sum(k_hops)
        perm = topk(score, self.pooling_ratio, batch)
        if perm == None:
            print(score)
        if ft:
            x = x[perm]
            batch = batch[perm]
            edge_index, edge_attr = filter_adj(edge_index, edge_attr, perm,
                                               num_nodes=score.size(0))

            return x, edge_index, edge_weight, batch, perm

        else:
            mask = torch.zeros_like(score)
            mask[perm] = 1
            x2 = x * mask.view(-1, 1)  # for these unselected nodes, set features to zero

            # edge_weights
            new_edge_weights = mask[edge_index[0]] + mask[edge_index[1]]
            edges_mask = torch.where(new_edge_weights != 2, torch.zeros(1, device=edge_index.device),
                                     torch.ones(1, device=edge_index.device))
            edge_weight2 = edge_weight * edges_mask

            return x2, edge_index, edge_weight2, batch, perm

class Gappool_Mixed(MessagePassing):
    def __init__(self, in_channels, ratio):
        super(Gappool_Mixed, self).__init__()
        self.pooling_ratio = ratio
        self.linear = torch.nn.Linear(in_channels, 1)
    def reset_parameters(self):
        self.linear.reset_parameters()
    def forward(self, x, edge_index, batch=None, edge_weight=None, edge_attr=None, ft=False):
        if batch is None:
            batch = edge_index.new_zeros(x.size(0))

        if edge_weight == None:
            edge_weight = torch.ones(edge_index.size()[1], device=x.device)
        # score = W \times sigma_{(x_u - x_j)^2} \times 0.5
        gap = self.propagate(edge_index, x=x, edge_weight=edge_weight)
        gap = self.linear(gap) / 2
        score = gap.reshape(gap.size(0))
        perm = topk(score, self.pooling_ratio, batch)
        if ft:
            x = x[perm] * score[perm].view(-1, 1)
            batch = batch[perm]
            edge_index, edge_attr = filter_adj(edge_index, edge_attr, perm,
                                               num_nodes=score.size(0))

            return x, edge_index, edge_weight, batch, perm
        else:
            mask = torch.zeros_like(score)
            mask[perm] = 1
            x1 = x * score.view(-1, 1)
            x2 = x1 * mask.view(-1, 1)  # for these unselected nodes, set features to zero

            # edge_weights
            new_edge_weights = mask[edge_index[0]] + mask[edge_index[1]]
            edges_mask = torch.where(new_edge_weights != 2, torch.zeros(1, device=edge_index.device),
                                     torch.ones(1, device=edge_index.device))
            edge_weight2 = edge_weight * edges_mask
            return x2, edge_index, edge_weight2, batch, perm

    def message(self, x_i, x_j, edge_weight):
        out = (x_i - x_j)*(x_i - x_j)
        return out if edge_weight is None else out * edge_weight.view(-1, 1)

def filter_features(x,edge_index, edge_weight, batch, th=0.001):
    score = torch.norm(x, p=1, dim=1)
    perm = topk(score, 0, batch, min_score=th)
    x = x[perm]
    batch = batch[perm]
    edge_index, edge_weight = filter_adj(edge_index, edge_weight, perm,
                                       num_nodes=score.size(0))
    # print('pool after mixed pooling: {},{}', x.size(0), edge_index.size(1))
    return x, edge_index, edge_weight, batch, perm

def filter_perm(x,edge_index, edge_weight, batch, perm_ori, th=0.001):
    perm = topk(perm_ori, 0, batch, min_score=th)
    x = x[perm]
    batch = batch[perm]
    edge_index, edge_weight = filter_adj(edge_index, edge_weight, perm,
                                       num_nodes=perm_ori.size(0))
    # print('pool after mixed pooling: {},{}', x.size(0), edge_index.size(1))
    return x, edge_index, edge_weight, batch, perm



from typing import Callable, Union
from torch_geometric.typing import OptPairTensor, Adj, OptTensor, Size

import torch
from torch import Tensor
import torch.nn.functional as F
from torch_sparse import SparseTensor, matmul
from torch_geometric.nn.conv import MessagePassing

from .inits import reset


class GINConv(MessagePassing):
    r"""The graph isomorphism operator from the `"How Powerful are
    Graph Neural Networks?" <https://arxiv.org/abs/1810.00826>`_ paper

    .. math::
        \mathbf{x}^{\prime}_i = h_{\mathbf{\Theta}} \left( (1 + \epsilon) \cdot
        \mathbf{x}_i + \sum_{j \in \mathcal{N}(i)} \mathbf{x}_j \right)

    or

    .. math::
        \mathbf{X}^{\prime} = h_{\mathbf{\Theta}} \left( \left( \mathbf{A} +
        (1 + \epsilon) \cdot \mathbf{I} \right) \cdot \mathbf{X} \right),

    here :math:`h_{\mathbf{\Theta}}` denotes a neural network, *.i.e.* an MLP.

    Args:
        nn (torch.nn.Module): A neural network :math:`h_{\mathbf{\Theta}}` that
            maps node features :obj:`x` of shape :obj:`[-1, in_channels]` to
            shape :obj:`[-1, out_channels]`, *e.g.*, defined by
            :class:`torch.nn.Sequential`.
        eps (float, optional): (Initial) :math:`\epsilon`-value.
            (default: :obj:`0.`)
        train_eps (bool, optional): If set to :obj:`True`, :math:`\epsilon`
            will be a trainable parameter. (default: :obj:`False`)
        **kwargs (optional): Additional arguments of
            :class:`torch_geometric.nn.conv.MessagePassing`.
    """

    def __init__(
        self, nn: Callable, eps: float = 0.0, train_eps: bool = False, **kwargs
    ):
        kwargs.setdefault("aggr", "add")
        super(GINConv, self).__init__(**kwargs)
        self.nn = nn
        self.initial_eps = eps
        if train_eps:
            self.eps = torch.nn.Parameter(torch.Tensor([eps]))
        else:
            self.register_buffer("eps", torch.Tensor([eps]))
        self.reset_parameters()

    def reset_parameters(self):
        reset(self.nn)
        self.eps.data.fill_(self.initial_eps)

    def forward(
        self,
        x: Union[Tensor, OptPairTensor],
        edge_index: Adj,
        edge_weight: OptTensor = None,
        size: Size = None,
    ) -> Tensor:
        """"""
        if isinstance(x, Tensor):
            x: OptPairTensor = (x, x)

        # propagate_type: (x: OptPairTensor)
        out = self.propagate(edge_index, x=x, edge_weight=edge_weight, size=size)

        x_r = x[1]
        if x_r is not None:
            out += (1 + self.eps) * x_r

        return self.nn(out)

    # def message(self, x_j: Tensor) -> Tensor:
    #    return x_j
    def message(self, x_j: Tensor, edge_weight: OptTensor) -> Tensor:
        if edge_weight is None:
            return x_j
        else:
            return edge_weight.view(-1, 1) * x_j

    def message_and_aggregate(self, adj_t: SparseTensor, x: OptPairTensor) -> Tensor:
        adj_t = adj_t.set_value(None, layout=None)
        return matmul(adj_t, x[0], reduce=self.aggr)

    def __repr__(self):
        return "{}(nn={})".format(self.__class__.__name__, self.nn)


class GINEConv(MessagePassing):
    r"""The modified :class:`GINConv` operator from the `"Strategies for
    Pre-training Graph Neural Networks" <https://arxiv.org/abs/1905.12265>`_
    paper

    .. math::
        \mathbf{x}^{\prime}_i = h_{\mathbf{\Theta}} \left( (1 + \epsilon) \cdot
        \mathbf{x}_i + \sum_{j \in \mathcal{N}(i)} \mathrm{ReLU}
        ( \mathbf{x}_j + \mathbf{e}_{j,i} ) \right)

    that is able to incorporate edge features :math:`\mathbf{e}_{j,i}` into
    the aggregation procedure.

    Args:
        nn (torch.nn.Module): A neural network :math:`h_{\mathbf{\Theta}}` that
            maps node features :obj:`x` of shape :obj:`[-1, in_channels]` to
            shape :obj:`[-1, out_channels]`, *e.g.*, defined by
            :class:`torch.nn.Sequential`.
        eps (float, optional): (Initial) :math:`\epsilon`-value.
            (default: :obj:`0.`)
        train_eps (bool, optional): If set to :obj:`True`, :math:`\epsilon`
            will be a trainable parameter. (default: :obj:`False`)
        **kwargs (optional): Additional arguments of
            :class:`torch_geometric.nn.conv.MessagePassing`.
    """

    def __init__(
        self, nn: Callable, eps: float = 0.0, train_eps: bool = False, **kwargs
    ):
        kwargs.setdefault("aggr", "add")
        super(GINEConv, self).__init__(**kwargs)
        self.nn = nn
        self.initial_eps = eps
        if train_eps:
            self.eps = torch.nn.Parameter(torch.Tensor([eps]))
        else:
            self.register_buffer("eps", torch.Tensor([eps]))
        self.reset_parameters()

    def reset_parameters(self):
        reset(self.nn)
        self.eps.data.fill_(self.initial_eps)

    def forward(
        self,
        x: Union[Tensor, OptPairTensor],
        edge_index: Adj,
        edge_attr: OptTensor = None,
        size: Size = None,
    ) -> Tensor:
        """"""
        if isinstance(x, Tensor):
            x: OptPairTensor = (x, x)

        # Node and edge feature dimensionalites need to match.
        if isinstance(edge_index, Tensor):
            assert edge_attr is not None
            assert x[0].size(-1) == edge_attr.size(-1)
        elif isinstance(edge_index, SparseTensor):
            assert x[0].size(-1) == edge_index.size(-1)

        # propagate_type: (x: OptPairTensor, edge_attr: OptTensor)
        out = self.propagate(edge_index, x=x, edge_attr=edge_attr, size=size)

        x_r = x[1]
        if x_r is not None:
            out += (1 + self.eps) * x_r

        return self.nn(out)

    def message(self, x_j: Tensor, edge_attr: Tensor) -> Tensor:
        return F.relu(x_j + edge_attr)

    def __repr__(self):
        return "{}(nn={})".format(self.__class__.__name__, self.nn)



import hyperopt

# from . import register_hpo
from .advisorbase import AdvisorBaseHPOptimizer
from .suggestion.algorithm.chocolate_bayes import ChocolateBayesAlgorithm

# @register_hpo("randadvisor")


class BayesAdvisorChoco(AdvisorBaseHPOptimizer):
    def __init__(self, args):
        super().__init__(args)
        self.method = ChocolateBayesAlgorithm()



from typing import Callable
from torch_geometric.typing import Adj, OptTensor

import torch
from torch import Tensor
from torch.nn import Parameter
import torch.nn.functional as F
from torch_sparse import SparseTensor, matmul
from torch_geometric.nn.conv import MessagePassing
from torch_geometric.nn.conv.gcn_conv import gcn_norm

from .inits import glorot, zeros


class ARMAConv(MessagePassing):
    r"""The ARMA graph convolutional operator from the `"Graph Neural Networks
    with Convolutional ARMA Filters" <https://arxiv.org/abs/1901.01343>`_ paper

    .. math::
        \mathbf{X}^{\prime} = \frac{1}{K} \sum_{k=1}^K \mathbf{X}_k^{(T)},

    with :math:`\mathbf{X}_k^{(T)}` being recursively defined by

    .. math::
        \mathbf{X}_k^{(t+1)} = \sigma \left( \mathbf{\hat{L}}
        \mathbf{X}_k^{(t)} \mathbf{W} + \mathbf{X}^{(0)} \mathbf{V} \right),

    where :math:`\mathbf{\hat{L}} = \mathbf{I} - \mathbf{L} = \mathbf{D}^{-1/2}
    \mathbf{A} \mathbf{D}^{-1/2}` denotes the
    modified Laplacian :math:`\mathbf{L} = \mathbf{I} - \mathbf{D}^{-1/2}
    \mathbf{A} \mathbf{D}^{-1/2}`.

    Args:
        in_channels (int): Size of each input sample :math:`\mathbf{x}^{(t)}`.
        out_channels (int): Size of each output sample
            :math:`\mathbf{x}^{(t+1)}`.
        num_stacks (int, optional): Number of parallel stacks :math:`K`.
            (default: :obj:`1`).
        num_layers (int, optional): Number of layers :math:`T`.
            (default: :obj:`1`)
        act (callable, optional): Activation function :math:`\sigma`.
            (default: :meth:`torch.nn.functional.ReLU`)
        shared_weights (int, optional): If set to :obj:`True` the layers in
            each stack will share the same parameters. (default: :obj:`False`)
        dropout (float, optional): Dropout probability of the skip connection.
            (default: :obj:`0.`)
        bias (bool, optional): If set to :obj:`False`, the layer will not learn
            an additive bias. (default: :obj:`True`)
        **kwargs (optional): Additional arguments of
            :class:`torch_geometric.nn.conv.MessagePassing`.
    """

    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        num_stacks: int = 1,
        num_layers: int = 1,
        shared_weights: bool = False,
        act: Callable = F.relu,
        normalize=True,
        dropout: float = 0.0,
        bias: bool = True,
        **kwargs
    ):
        kwargs.setdefault("aggr", "add")
        super(ARMAConv, self).__init__(**kwargs)

        self.in_channels = in_channels
        self.out_channels = out_channels
        self.num_stacks = num_stacks
        self.num_layers = num_layers
        self.act = act
        self.shared_weights = shared_weights
        self.dropout = dropout
        self.normalize = normalize

        K, T, F_in, F_out = num_stacks, num_layers, in_channels, out_channels
        T = 1 if self.shared_weights else T

        self.init_weight = Parameter(torch.Tensor(K, F_in, F_out))
        self.weight = Parameter(torch.Tensor(max(1, T - 1), K, F_out, F_out))
        self.root_weight = Parameter(torch.Tensor(T, K, F_in, F_out))

        if bias:
            self.bias = Parameter(torch.Tensor(T, K, 1, F_out))
        else:
            self.register_parameter("bias", None)

        self.reset_parameters()

    def reset_parameters(self):
        glorot(self.init_weight)
        glorot(self.weight)
        glorot(self.root_weight)
        zeros(self.bias)

    def forward(
        self, x: Tensor, edge_index: Adj, edge_weight: OptTensor = None
    ) -> Tensor:
        """"""

        if isinstance(edge_index, Tensor) and self.normalize:
            edge_index, edge_weight = gcn_norm(  # yapf: disable
                edge_index,
                edge_weight,
                x.size(self.node_dim),
                add_self_loops=False,
                dtype=x.dtype,
            )

        elif isinstance(edge_index, SparseTensor) and self.normalize:
            edge_index = gcn_norm(  # yapf: disable
                edge_index,
                edge_weight,
                x.size(self.node_dim),
                add_self_loops=False,
                dtype=x.dtype,
            )

        x = x.unsqueeze(-3)
        out = x
        for t in range(self.num_layers):
            if t == 0:
                out = out @ self.init_weight
            else:
                out = out @ self.weight[0 if self.shared_weights else t - 1]

            # propagate_type: (x: Tensor, edge_weight: OptTensor)
            out = self.propagate(edge_index, x=out, edge_weight=edge_weight, size=None)

            root = F.dropout(x, p=self.dropout, training=self.training)
            out += root @ self.root_weight[0 if self.shared_weights else t]

            if self.bias is not None:
                out += self.bias[0 if self.shared_weights else t]

            out = self.act(out)

        return out.mean(dim=-3)

    def message(self, x_j: Tensor, edge_weight: Tensor) -> Tensor:
        return edge_weight.view(-1, 1) * x_j

    def message_and_aggregate(self, adj_t: SparseTensor, x: Tensor) -> Tensor:
        return matmul(adj_t, x, reduce=self.aggr)

    def __repr__(self):
        return "{}({}, {}, num_stacks={}, num_layers={})".format(
            self.__class__.__name__,
            self.in_channels,
            self.out_channels,
            self.num_stacks,
            self.num_layers,
        )



from .operation import gnn_map
import typing as _typ
import torch

import torch.nn.functional as F

from .base import BaseSpace
from ...utils.backend import BackendOperator as BK


class SinglePathNodeClassificationSpace(BaseSpace):
    def __init__(
        self,
        hidden_dim: _typ.Optional[int] = 64,
        layer_number: _typ.Optional[int] = 2,
        dropout: _typ.Optional[float] = 0.2,
        input_dim: _typ.Optional[int] = None,
        output_dim: _typ.Optional[int] = None,
        ops: _typ.Tuple = ["gcn", "gat_8"],
    ):
        super().__init__()
        self.layer_number = layer_number
        self.hidden_dim = hidden_dim
        self.input_dim = input_dim
        self.output_dim = output_dim
        self.ops = ops
        self.dropout = dropout

    def build_graph(self):
        for layer in range(self.layer_number):
            key = f"op_{layer}"
            in_dim = self.input_dim if layer == 0 else self.hidden_dim
            out_dim = (
                self.output_dim if layer == self.layer_number - 1 else self.hidden_dim
            )
            op_candidates = [
                op(in_dim, out_dim)
                if isinstance(op, type)
                else gnn_map(op, in_dim, out_dim)
                for op in self.ops
            ]
            self.setLayerChoice(layer, op_candidates, key=key)

    def forward(self, data):
        x = BK.feat(data)
        for layer in range(self.layer_number):
            op = getattr(self, f"op_{layer}")
            x = BK.gconv(op, data, x)
            if layer != self.layer_number - 1:
                x = F.leaky_relu(x)
                x = F.dropout(x, p=self.dropout, training=self.training)
        return F.log_softmax(x, dim=1)



import json
import chocolate as choco

from ...suggestion.models import Study
from ...suggestion.models import Trial
from ...suggestion.algorithm.abstract_algorithm import AbstractSuggestionAlgorithm


class BaseChocolateAlgorithm(AbstractSuggestionAlgorithm):
    """
    The implementation is based on https://github.com/tobegit3hub/advisor
    """

    def __init__(self, algorithm_name="QuasiRandom"):

        self.algorithm_name = algorithm_name

    def get_new_suggestions(self, study, trials=[], number=1):
        """
        Get the new suggested trials with Chocolate algorithm.
        """

        # 1. Construct search space
        # Example: {"x" : choco.uniform(-6, 6), "y" : choco.uniform(-6, 6)}
        chocolate_search_space = {}

        # study = Study.objects.get(name=study_name)
        study_configuration_json = json.loads(study.study_configuration)
        params = study_configuration_json["params"]

        for param in params:
            param_name = param["parameterName"]

            if param["type"] == "INTEGER":
                # TODO: Support int type of search space)
                pass

            elif param["type"] == "DOUBLE":
                chocolate_search_space[param_name] = choco.uniform(
                    param["minValue"], param["maxValue"]
                )

            elif param["type"] == "DISCRETE" or param["type"] == "CATEGORICAL":
                feasible_point_list = [
                    value.strip() for value in param["feasiblePoints"].split(",")
                ]
                chocolate_search_space[param_name] = choco.choice(feasible_point_list)

        conn = choco.SQLiteConnection("sqlite:///my_db.db")

        # Refer to https://chocolate.readthedocs.io/tutorials/algo.html
        if self.algorithm_name == "Grid":
            sampler = choco.Grid(conn, chocolate_search_space, clear_db=True)
        elif self.algorithm_name == "Random":
            sampler = choco.Random(conn, chocolate_search_space, clear_db=True)
        elif self.algorithm_name == "QuasiRandom":
            sampler = choco.QuasiRandom(conn, chocolate_search_space, clear_db=True)
        elif self.algorithm_name == "Bayes":
            sampler = choco.Bayes(conn, chocolate_search_space, clear_db=True)
        elif self.algorithm_name == "CMAES":
            sampler = choco.CMAES(conn, chocolate_search_space, clear_db=True)
        elif self.algorithm_name == "MOCMAES":
            mu = 1
            sampler = choco.MOCMAES(conn, chocolate_search_space, mu=mu, clear_db=True)

        # 2. Update with completed advisor trials
        # completed_advisor_trials = Trial.objects.filter(
        #    study_name=study_name, status="Completed")
        completed_advisor_trials = [i for i in trials if i.status == "Completed"]

        for index, advisor_trial in enumerate(completed_advisor_trials):
            parameter_values_json = json.loads(advisor_trial.parameter_values)

            loss = advisor_trial.objective_value
            if study_configuration_json["goal"] == "MAXIMIZE":
                loss = -1 * loss

            entry = {"_chocolate_id": index, "_loss": loss}
            entry.update(parameter_values_json)
            # Should not use sampler.update(token, loss)
            conn.insert_result(entry)

        # 3. Run algorithm and construct return advisor trials
        return_trial_list = []

        for i in range(number):

            # Example: {'_chocolate_id': 1}
            # Example: {u'hidden2': u'32', u'learning_rate': 0.07122424534644338, u'l1_normalization': 0.8402644688674471, u'optimizer': u'adam'}
            token, chocolate_params = sampler.next()

            parameter_values_json = {}

            for param in params:

                if (
                    param["type"] == "INTEGER"
                    or param["type"] == "DOUBLE"
                    or param["type"] == "CATEGORICAL"
                ):
                    parameter_values_json[param["parameterName"]] = chocolate_params[
                        param["parameterName"]
                    ]
                elif param["type"] == "DISCRETE":
                    parameter_values_json[param["parameterName"]] = int(
                        chocolate_params[param["parameterName"]]
                    )

            new_advisor_trial = Trial.create(study.name, "ChocolateTrial")
            new_advisor_trial.parameter_values = json.dumps(parameter_values_json)
            # new_advisor_trial.save()
            return_trial_list.append(new_advisor_trial)

        return return_trial_list



import dataclasses
import random
import numpy as np


@dataclasses.dataclass
class Individual:
    """
    A class that represents an individual.
    Holds two attributes, where ``x`` is the model and ``y`` is the metric (e.g., accuracy).
    """

    x: dict
    y: float


class MutationSampler:
    """uniform mutator

    Parameters
    ----------
    nas_modules: 
        nas_modules in NAS algorithms , including choices of modules
    mutation_prob: float
        probability of doing mutation in each choice.
    parent : dict
        parent individual's choices
    """

    def __init__(self, nas_modules, mutation_prob):
        selection_range = {}
        for k, v in nas_modules:
            selection_range[k] = len(v)
        self.selection_dict = selection_range
        self.mutation_prob = mutation_prob

    def resample(self, parent):
        search_space = self.selection_dict
        child = {}
        for k, v in parent.items():
            if random.uniform(0, 1) < self.mutation_prob:
                child[k] = np.random.choice(
                    range(search_space[k])
                )  # do not exclude the original operator
            else:
                child[k] = v
        return child


class UniformSampler:
    """Uniform Sampler

    Parameters
    ----------
    nas_modules: 
        nas_modules in NAS algorithms , including choices of modules
    """

    def __init__(self, nas_modules):
        selection_range = {}
        for k, v in nas_modules:
            selection_range[k] = len(v)
        self.selection_dict = selection_range

    def resample(self):
        selection = {}
        for k, v in self.selection_dict.items():
            selection[k] = np.random.choice(range(v))
        return selection



import torch
import torch.nn as nn
import torch.nn.functional as F

class AGLayer(nn.Module):
    def __init__(self, args, num_op):
        super().__init__()
        self.args = args
        self.op_emb = nn.Embedding(num_op, args.graph_dim) # op * g_d

    def forward(self, g):
        # g: graph * g_d
        o = self.op_emb.weight
        o = o / o.norm(2, dim = -1, keepdim = True)
        cosloss = (o @ o.t()).sum()
        alpha = g @ o.t()
        alpha = alpha / self.args.temperature
        alpha = F.softmax(alpha, dim = 1) # graph * op
        alpha = alpha * (alpha > 1/6)
        alpha = alpha / alpha.sum(dim = 1, keepdim = True)
        return alpha, cosloss

class AG(nn.Module):
    def __init__(self, args, num_op, num_pool):
        super().__init__()
        self.args = args
        self.layers = nn.ModuleList()
        self.set = 'train'
        for i in range(args.num_layers + 1):
            self.layers.append(AGLayer(args, num_op))

    def forward(self, g):
        # g: graph * g_d
        alpha_all = []
        cosloss = torch.zeros(1).to(self.layers[0].op_emb.weight.device)

        for i in range(self.args.num_layers):
            alpha, closs = self.layers[i](g)
            cosloss = cosloss + closs
            alpha_all.append(alpha)
        #alpha_all = [i.detach() for i in alpha_all]

        return alpha_all, cosloss



"""
HPO Module for tuning hyper parameters
"""

import math
import multiprocessing as mp
import random
import time
from abc import ABC, abstractmethod

from .suggestion.models import Trial


class HPSpace(ABC):
    def __init__(self, name):
        self.name = name

    @abstractmethod
    def to_config(self):
        pass


class RangeHP(HPSpace):
    def __init__(self, name, min_value, max_value):
        super.__init__(name)
        self.min = min_value
        self.max = max_value

    def to_config(self):
        return {
            "parameterName": self.name,
            "type": "DOUBLE",
            "minValue": self.min,
            "maxValue": self.max,
            "scalingType": "LINEAR",
        }


class LogRangeHP(HPSpace):
    def __init__(self, name, min_value, max_value):
        super().__init__(name)
        self.min = min_value
        self.max = max_value

    def to_config(self):
        return {
            "parameterName": self.name,
            "type": "DOUBLE",
            "minValue": self.min,
            "maxValue": self.max,
            "scalingType": "LOG",
        }


class ChoiceHP(HPSpace):
    def __init__(self, name, choices):
        super().__init__(name)
        self.choices = choices

    def to_config(self):
        return {
            "parameterName": self.name,
            "type": "CATEGORICAL",
            "feasiblePoints": self.choices,
        }


class BaseHPOptimizer:
    def __init__(self, hp_space, f, time_limit=None, memory_limit=None, max_evals=None):
        self.hp_space = self._hpspace_to_config(hp_space)
        self.f = f
        self.trials = []
        self.time_limit = time_limit
        self.memory_limit = memory_limit
        self.max_evals = max_evals

    def _hpspace_to_config(self, hp_space):
        tmp = []
        for hp in hp_space:
            tmp.append(hp.to_config())
        return {"space": tmp}

    def _decompose_dld(self, config):
        self._dld = {}
        if isinstance(config, list):
            return config
        # config is dict
        list_config = []
        for key in config:
            self._dld[key] = []
            for para in config[key]:
                self._dld[key].append(para["parameterName"])
                newpara = para.copy()
                newpara["parameterName"] = key + ":" + para["parameterName"]
                if "cutPara" in para.keys():
                    if isinstance(newpara["cutPara"], str):
                        newpara["cutPara"] = key + ":" + para["cutPara"]
                    else:
                        newpara["cutPara"] = [
                            key + ":" + cutname for cutname in para["cutPara"]
                        ]
                list_config.append(newpara)
        return list_config

    def _compose_dld(self, para):
        if not self._dld:
            return para
        fin = {}
        for key in self._dld:
            fin[key] = {}
            for pname in self._dld[key]:
                fin[key][pname] = para[key + ":" + pname]
        return fin

    def _decompose_depend_list_para(self, config):
        self._depend_map = {}

        def get_depended_para(name):
            for p in config:
                if p["parameterName"] == name:
                    return
            raise WrongDependedParameterError("The depended parameter does not exist.")

        for para in config:
            if para["type"] in ("NUMERICAL_LIST", "CATEGORICAL_LIST") and para.get(
                "cutPara", None
            ):
                self._depend_map[para["parameterName"]] = para
                if type(para["cutPara"]) == str:
                    get_depended_para(para["cutPara"])
                else:
                    for dpara in para["cutPara"]:
                        get_depended_para(dpara)

        return config

    def _compose_depend_list_para(self, config):
        for para in self._depend_map:
            cutparas = self._depend_map[para]["cutPara"]
            if type(cutparas) == str:
                dparas = [config[cutparas]]
                # dparas = config[cutparas]
            else:
                dparas = []
                for dpara in cutparas:
                    dparas.append(config[dpara])
            paralen = self._depend_map[para]["cutFunc"](dparas)
            config[para] = config[para][:paralen]
        return config

    def _decompose_list_fixed_para(self, config):
        fin = []
        self._list_map = {}
        self._fix_map = {}
        for para in config:
            if para["type"] == "NUMERICAL_LIST":
                self._list_map[para["parameterName"]] = para["length"]
                if type(para["minValue"]) != list:
                    para["minValue"] = [para["minValue"] for i in range(para["length"])]
                if type(para["maxValue"]) != list:
                    para["maxValue"] = [para["maxValue"] for i in range(para["length"])]
                for i, x, y in zip(
                    range(para["length"]), para["minValue"], para["maxValue"]
                ):
                    new_para = {}
                    new_para["parameterName"] = para["parameterName"] + "_" + str(i)
                    new_para["type"] = para["numericalType"]
                    new_para["minValue"] = x
                    new_para["maxValue"] = y
                    new_para["scalingType"] = para["scalingType"]
                    fin.append(new_para)
            elif para["type"] == "CATEGORICAL_LIST":
                self._list_map[para["parameterName"]] = para["length"]
                category = para["feasiblePoints"]
                self._category_map[para["parameterName"]] = category

                cur_points = ",".join(map(lambda _x: str(_x), range(len(category))))
                for i in range(para["length"]):
                    new_para = dict()
                    new_para["parameterName"] = para["parameterName"] + "_" + str(i)
                    new_para["type"] = "DISCRETE"
                    new_para["feasiblePoints"] = cur_points
                    fin.append(new_para)
            elif para["type"] == "FIXED":
                self._fix_map[para["parameterName"]] = para["value"]
            else:
                new_para = para.copy()
                new_para["parameterName"] = para["parameterName"] + "_"
                fin.append(new_para)
        return fin

    def _compose_list_fixed_para(self, config):
        fin = {}
        # compose list
        for pname in self._list_map:
            val = []
            for i in range(self._list_map[pname]):
                val.append(config[pname + "_" + str(i)])
                del config[pname + "_" + str(i)]
            if pname in self._category_map:
                val = [self._category_map[pname][i] for i in val]
            fin[pname] = val
        # deal other para
        for pname in config:
            fin[pname[:-1]] = config[pname]
        for pname in self._fix_map:
            fin[pname] = self._fix_map[pname]
        return fin

    def _encode_para_convert(self, config):
        """
        Convert all types of para space to DOUBLE(linear), DISCRETE
        config: [{
            "parameterName": "num_layers",
            "type": "DISCRETE",
            "feasiblePoints": "1,2,3,4",
        },{
            "parameterName": "hidden",
            "type": "NUMERICAL_LIST",
            "numericalType": "INTEGER",
            "length": 4,
            "minValue": [4, 4, 4, 4],
            "maxValue": [32, 32, 32, 32],
            "scalingType": "LOG"
        },{
            "parameterName": "dropout",
            "type": "DOUBLE",
            "minValue": 0.1,
            "maxValue": 0.9,
            "scalingType": "LINEAR"
        }]"""
        self._category_map = {}
        self._discrete_map = {}
        self._numerical_map = {}

        current_config = []
        for para in config:
            if para["type"] == "DOUBLE" or para["type"] == "INTEGER":
                cur_para = para.copy()
                cur_para["type"] = "DOUBLE"
                if para["scalingType"] == "LOG":
                    cur_para["minValue"] = math.log(para["minValue"])
                    cur_para["maxValue"] = math.log(para["maxValue"])
                current_config.append(cur_para)
                self._numerical_map[para["parameterName"]] = para
            elif para["type"] == "CATEGORICAL" or para["type"] == "DISCRETE":
                if para["type"] == "DISCRETE":
                    cate_list = para["feasiblePoints"].split(",")
                    cate_list = list(map(lambda x: x.strip(), cate_list))
                else:
                    cate_list = para["feasiblePoints"]
                cur_points = ",".join(map(lambda x: str(x), range(len(cate_list))))
                cur_para = para.copy()
                cur_para["feasiblePoints"] = cur_points
                cur_para["type"] = "DISCRETE"
                current_config.append(cur_para)
                if para["type"] == "CATEGORICAL":
                    self._category_map[para["parameterName"]] = cate_list
                else:
                    self._discrete_map[para["parameterName"]] = cate_list
            else:
                current_config.append(para)
        return current_config

    def _decode_para_convert(self, para):
        """
        decode HPO given para to user(externel) para and trial para
        """
        externel_para = para.copy()
        trial_para = para.copy()
        for name in para:
            if name in self._numerical_map:
                old_para = self._numerical_map[name]
                val = para[name]
                if old_para["scalingType"] == "LOG":
                    val = math.exp(val)
                    if val < old_para["minValue"]:
                        val = old_para["minValue"]
                    elif val > old_para["maxValue"]:
                        val = old_para["maxValue"]
                if old_para["type"] == "INTEGER":
                    val = int(round(val))
                externel_para[name] = val
                trial_para[name] = (
                    val if old_para["scalingType"] != "LOG" else math.log(val)
                )
            elif name in self._category_map:
                externel_para[name] = self._category_map[name][int(para[name])]
                trial_para[name] = para[name]
            elif name in self._discrete_map:
                externel_para[name] = eval(self._discrete_map[name][int(para[name])])
                trial_para[name] = para[name]
        return externel_para, trial_para

    def _encode_para(self, config):
        # dict(list(dict)) -> list(dict)
        config = self._decompose_dld(config)
        # (dependent list) -> list
        config = self._decompose_depend_list_para(config)
        # list -> double/discrete; fixed -> removed
        config = self._decompose_list_fixed_para(config)
        # discrete -> categorical; log -> linear
        config = self._encode_para_convert(config)
        return config

    def _decode_para(self, para):
        para, trial_para = self._decode_para_convert(para)
        para = self._compose_list_fixed_para(para)
        para = self._compose_depend_list_para(para)
        para = self._compose_dld(para)
        return para, trial_para

    def slave(self, pipe):
        while 1:
            x = pipe.recv()
            y = self.f(x)
            pipe.send((x, y))

    def _set_up(self, num_slaves, time_limit, memory_limit):
        """
        Initialize something used in "optimize"

        Parameters
        ----------
        trainer : ..train.BaseTrainer
            Including model, giving HP space and using for training
        dataset : ...datasets
            Dataset to train and evaluate.
        time_limit : int
            Max time to run HPO
        memory_limit : None
            No implementation yet
        """
        pass

    def _update_trials(self, pid, hp, perf):
        """
        After the evaluation phase of each turn, update history trials according to the performance

        Parameters
        ----------
        pid : int
            The process id which runs the evaluation
        hp : dict
            The HPs used in evaluation
        perf : float
            The performance of the HP, higher is better
        """
        pass

    def _get_suggestion(self, pid):
        """
        Give the next HP suggestion

        Parameters
        ----------
        pid : int
            The process id which will run the evaluation

        Returns
        -------
        para_json: dict
            The suggested HP
        """
        pass

    def _best_hp(self):
        """
        Give the best HP and the best trainer as the returns of "optimize"

        Returns
        -------
        trainer: ..train.BaseTrainer
            The trainer including the best trained model
        para_json: dict
            The best HP
        """
        pass

    def optimize(self):
        """Key function. Return the best hp & performance"""
        start_time = time.time()
        cur_evals = 0

        slaves = 2
        pipes = []
        procs = []
        statuses = []
        for i in range(slaves):
            ps, pr = mp.Pipe()
            pipes.append(ps)
            proc = mp.Process(
                target=self.slave,
                args=(pr,),
            )
            proc.start()
            procs.append(proc)
            statuses.append("ready")

        # do something of the certain HPO algo
        self._set_up(slaves, self.time_limit, self.memory_limit)

        while cur_evals < self.max_evals:
            # timeout
            if time.time() - start_time > self.time_limit:
                self.logger.info("Time out of limit, Epoch: {}".format(str(cur_evals)))
                break
            for i in range(slaves):
                sent = pipes[i].poll()
                if sent or statuses[i] == "ready":
                    if sent:
                        x, y = pipes[i].recv()
                        self.trials.append(
                            self._creat_a_trail("HPO", "Base", "Completed", x, y)
                        )
                        self._update_trials(i, x, y)
                        cur_evals += 1
                        if cur_evals >= self.max_evals:
                            break
                    else:
                        statuses[i] = "running"
                    x = self._get_suggestion(i)
                    pipes[i].send(x)

        for i in range(slaves):
            procs[i].terminate()

        best_x, best_perf = self._best_hp()
        if best_x == None:
            raise TimeTooLimitedError(
                "Given time is too limited to finish one round in HPO."
            )
        return best_x, best_perf

    def _creat_a_trail(
        self, study_name, name, status, parametrer_values, objective_value
    ):
        trial = Trial.create(study_name, name)
        trial.status = status
        trial.parameter_values = parametrer_values
        trial.objective_value = objective_value
        return trial

    def all_trials(self):
        raise self.trials


class TimeTooLimitedError(Exception):
    pass


class WrongDependedParameterError(Exception):
    pass


def f(hps: dict, dataset=None):
    # build a module with hps
    # if dataset is not None, use the provided dataset, or use the dataset from trainer otherwise
    model = get_model(hps["num_layers"])
    trainer = Trainer(data)
    trainer.set_hp(hps)
    trainer.train(model)
    return trainer.evaluate(model)


def test():

    space = [
        LogRangeHP("lr", 0.001, 0.1),
        ChoiceHP("opti", ["adam", "sgd"]),
        ChoiceHP("num_layers", [2, 3]),
    ]
    bhpo = BaseHPOptimizer(space, f)
    bhpo.optimize()


if __name__ == "__main__":
    test()



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT license.

import logging
from collections import OrderedDict

import time
import numpy as np
import torch


def get_hardware_aware_metric(model, hardware_metric):
    """
    Get architectures' hardware-aware metrics

    Attributes
    ----------
    model : BaseSpace
        The architecture to be evaluated
    hardware_metric : str
        The name of hardware-aware metric. Can be 'parameter' or 'latency'
    """

    if hardware_metric == "parameter":
        return count_parameters(model)
    elif hardware_metric == "latency":
        return measure_latency(model, 20, warmup_iters=5)
    else:
        raise ValueError("Unsupported hardware-aware metric")


def count_parameters(module, only_trainable=False):
    s = sum(
        p.numel()
        for p in module.parameters(recurse=False)
        if not only_trainable or p.requires_grad
    )
    if isinstance(module, PathSamplingLayerChoice):
        s += sum(count_parameters(m) for m in module.sampled_choices())
    else:
        s += sum(count_parameters(m) for m in module.children())
    return s


def measure_latency(model, num_iters=200, *, warmup_iters=50):
    device = next(model.parameters()).device
    num_feat = model.input_dim
    model.eval()
    latencys = []
    data = _build_random_data(device, num_feat)
    with torch.no_grad():
        try:
            for i in range(warmup_iters + num_iters):
                if device.type == "cuda":
                    torch.cuda.synchronize()
                start = time.time()
                model(data)
                if device.type == "cuda":
                    torch.cuda.synchronize()
                dt = time.time() - start
                if i >= warmup_iters:
                    latencys.append(dt)
        except RuntimeError as e:
            if "cuda" in str(e) or "CUDA" in str(e):
                INF = 100
                return INF
            else:
                raise e

    return np.mean(latencys)


def _build_random_data(device, num_feat):
    node_nums = 3000
    edge_nums = 10000

    class Data:
        pass

    data = Data()
    data.x = torch.randn((node_nums, num_feat)).to(device)
    data.edge_index = torch.randint(0, node_nums, (2, edge_nums)).to(device)
    data.num_features = num_feat
    return data


def to_device(obj, device):
    """
    Move a tensor, tuple, list, or dict onto device.
    """
    if torch.is_tensor(obj):
        return obj.to(device)
    if isinstance(obj, tuple):
        return tuple(to_device(t, device) for t in obj)
    if isinstance(obj, list):
        return [to_device(t, device) for t in obj]
    if isinstance(obj, dict):
        return {k: to_device(v, device) for k, v in obj.items()}
    if isinstance(obj, (int, float, str)):
        return obj
    raise ValueError("'%s' has unsupported type '%s'" % (obj, type(obj)))


def to_list(arr):
    if torch.is_tensor(arr):
        return arr.cpu().numpy().tolist()
    if isinstance(arr, np.ndarray):
        return arr.tolist()
    if isinstance(arr, (list, tuple)):
        return list(arr)
    return arr


class AverageMeterGroup:
    """
    Average meter group for multiple average meters.
    """

    def __init__(self):
        self.meters = OrderedDict()

    def update(self, data):
        """
        Update the meter group with a dict of metrics.
        Non-exist average meters will be automatically created.
        """
        for k, v in data.items():
            if k not in self.meters:
                self.meters[k] = AverageMeter(k, ":4f")
            self.meters[k].update(v)

    def __getattr__(self, item):
        return self.meters[item]

    def __getitem__(self, item):
        return self.meters[item]

    def __str__(self):
        return "  ".join(str(v) for v in self.meters.values())

    def summary(self):
        """
        Return a summary string of group data.
        """
        return "  ".join(v.summary() for v in self.meters.values())


class AverageMeter:
    """
    Computes and stores the average and current value.

    Parameters
    ----------
    name : str
        Name to display.
    fmt : str
        Format string to print the values.
    """

    def __init__(self, name, fmt=":f"):
        self.name = name
        self.fmt = fmt
        self.reset()

    def reset(self):
        """
        Reset the meter.
        """
        self.val = 0
        self.avg = 0
        self.sum = 0
        self.count = 0

    def update(self, val, n=1):
        """
        Update with value and weight.

        Parameters
        ----------
        val : float or int
            The new value to be accounted in.
        n : int
            The weight of the new value.
        """
        self.val = val
        self.sum += val * n
        self.count += n
        self.avg = self.sum / self.count

    def __str__(self):
        fmtstr = "{name} {val" + self.fmt + "} ({avg" + self.fmt + "})"
        return fmtstr.format(**self.__dict__)

    def summary(self):
        fmtstr = "{name}: {avg" + self.fmt + "}"
        return fmtstr.format(**self.__dict__)



import numpy as np
import typing as _typing
from sklearn.metrics import (
    f1_score,
    log_loss,
    accuracy_score,
    roc_auc_score,
    label_ranking_average_precision_score,
)


class Evaluation:
    @staticmethod
    def get_eval_name() -> str:
        """ Expected to return the name of this evaluation method """
        raise NotImplementedError

    @staticmethod
    def is_higher_better() -> bool:
        """ Expected to return whether this evaluation method is higher better (bool) """
        raise NotImplementedError

    @staticmethod
    def evaluate(predict, label) -> float:
        """ Expected to return the evaluation result (float) """
        raise NotImplementedError


class EvaluatorUtility:
    """ Auxiliary utilities for evaluation """

    class PredictionBatchCumulativeBuilder:
        """
        Batch-cumulative builder for prediction
        For large graph, as it is infeasible to predict all the nodes
        in validation set and test set in single batch,
        and layer-wise prediction mechanism is a practical evaluation approach,
        a batch-cumulative prediction collector `PredictionBatchCumulativeBuilder`
        is implemented for prediction in mini-batch manner.
        """

        def __init__(self):
            self.__indexes_in_integral_data: _typing.Optional[np.ndarray] = None
            self.__prediction: _typing.Optional[np.ndarray] = None

        def clear_batches(
            self, *__args, **__kwargs
        ) -> "EvaluatorUtility.PredictionBatchCumulativeBuilder":
            self.__indexes_in_integral_data = None
            self.__prediction = None
            return self

        def add_batch(
            self, indexes_in_integral_data: np.ndarray, batch_prediction: np.ndarray
        ) -> "EvaluatorUtility.PredictionBatchCumulativeBuilder":
            if not (
                isinstance(indexes_in_integral_data, np.ndarray)
                and isinstance(batch_prediction, np.ndarray)
                and len(indexes_in_integral_data.shape) == 1
            ):
                raise TypeError
            elif indexes_in_integral_data.shape[0] != batch_prediction.shape[0]:
                raise ValueError

            if self.__indexes_in_integral_data is None:
                if (
                    indexes_in_integral_data.shape
                    != np.unique(indexes_in_integral_data).shape
                ):
                    raise ValueError(
                        f"There exists duplicate index "
                        f"in the argument indexes_in_integral_data {indexes_in_integral_data}"
                    )
                else:
                    self.__indexes_in_integral_data: np.ndarray = np.unique(
                        indexes_in_integral_data
                    )
            else:
                __indexes_in_integral_data = np.concatenate(
                    (self.__indexes_in_integral_data, indexes_in_integral_data)
                )
                if (
                    __indexes_in_integral_data.shape
                    != np.unique(__indexes_in_integral_data).shape
                ):
                    raise ValueError
                else:
                    self.__indexes_in_integral_data: np.ndarray = (
                        __indexes_in_integral_data
                    )

            if self.__prediction is None:
                self.__prediction: np.ndarray = batch_prediction
            else:
                self.__prediction: np.ndarray = np.concatenate(
                    (self.__prediction, batch_prediction)
                )

            return self

        def compose(
            self, __sorted: bool = True, **__kwargs
        ) -> _typing.Tuple[np.ndarray, np.ndarray]:
            if __sorted:
                sorted_index = np.argsort(self.__indexes_in_integral_data)
                return (
                    self.__indexes_in_integral_data[sorted_index],
                    self.__prediction[sorted_index],
                )
            else:
                return self.__indexes_in_integral_data, self.__prediction


EVALUATE_DICT: _typing.Dict[str, _typing.Type[Evaluation]] = {}


def register_evaluate(*name):
    def register_evaluate_cls(cls):
        for n in name:
            if n in EVALUATE_DICT:
                raise ValueError("Cannot register duplicate evaluator ({})".format(n))
            if not issubclass(cls, Evaluation):
                raise ValueError(
                    "Evaluator ({}: {}) must extend Evaluation".format(n, cls.__name__)
                )
            EVALUATE_DICT[n] = cls
        return cls

    return register_evaluate_cls


def get_feval(feval):
    if isinstance(feval, str):
        return EVALUATE_DICT[feval]
    if isinstance(feval, type) and issubclass(feval, Evaluation):
        return feval
    if isinstance(feval, _typing.Sequence):
        return [get_feval(f) for f in feval]
    raise ValueError("feval argument of type", type(feval), "is not supported!")


class EvaluationUniversalRegistry:
    @classmethod
    def register_evaluation(
        cls, *names
    ) -> _typing.Callable[[_typing.Type[Evaluation]], _typing.Type[Evaluation]]:
        def _register_evaluation(
            _class: _typing.Type[Evaluation],
        ) -> _typing.Type[Evaluation]:
            for n in names:
                if n in EVALUATE_DICT:
                    raise ValueError(
                        "Cannot register duplicate evaluator ({})".format(n)
                    )
                if not issubclass(_class, Evaluation):
                    raise ValueError(
                        "Evaluator ({}: {}) must extend Evaluation".format(
                            n, cls.__name__
                        )
                    )
                EVALUATE_DICT[n] = _class
            return _class

        return _register_evaluation


@register_evaluate("logloss")
class Logloss(Evaluation):
    @staticmethod
    def get_eval_name():
        return "logloss"

    @staticmethod
    def is_higher_better():
        """
        Should return whether this evaluation method is higher better (bool)
        """
        return False

    @staticmethod
    def evaluate(predict, label):
        """
        Should return: the evaluation result (float)
        """
        return log_loss(label, predict)


@register_evaluate("auc", "ROC-AUC")
class Auc(Evaluation):
    @staticmethod
    def get_eval_name():
        return "auc"

    @staticmethod
    def is_higher_better():
        """
        Should return whether this evaluation method is higher better (bool)
        """
        return True

    @staticmethod
    def evaluate(predict, label):
        """
        Should return: the evaluation result (float)
        """
        if len(predict.shape) == 1:
            pos_predict = predict
        else:
            assert (
                predict.shape[1] == 2
            ), "Cannot use auc on given data with %d classes!" % (predict.shape[1])
            pos_predict = predict[:, 1]
        return roc_auc_score(label, pos_predict)


@register_evaluate("acc", "Accuracy")
class Acc(Evaluation):
    @staticmethod
    def get_eval_name():
        return "acc"

    @staticmethod
    def is_higher_better():
        """
        Should return whether this evaluation method is higher better (bool)
        """
        return True

    @staticmethod
    def evaluate(predict, label):
        """
        Should return: the evaluation result (float)
        """
        if len(predict.shape) == 2:
            predict = np.argmax(predict, axis=1)
        else:
            predict = [1 if p > 0.5 else 0 for p in predict]
        return accuracy_score(label, predict)


@register_evaluate("mrr")
class Mrr(Evaluation):
    @staticmethod
    def get_eval_name():
        return "mrr"

    @staticmethod
    def is_higher_better():
        """
        Should return whether this evaluation method is higher better (bool)
        """
        return True

    @staticmethod
    def evaluate(predict, label):
        """
        Should return: the evaluation result (float)
        """
        if len(predict.shape) == 2:
            assert (
                predict.shape[1] == 2
            ), "Cannot use mrr on given data with %d classes!" % (predict.shape[1])
            pos_predict = predict[:, 1]
        else:
            pos_predict = predict
        return label_ranking_average_precision_score(label, pos_predict)


@register_evaluate("MicroF1")
class MicroF1(Evaluation):
    @staticmethod
    def get_eval_name() -> str:
        return "MicroF1"

    @staticmethod
    def is_higher_better() -> bool:
        return True

    @staticmethod
    def evaluate(predict, label) -> float:
        return f1_score(label, np.argmax(predict, axis=1), average="micro")



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

# from django.db import models


class Study:
    # name = models.CharField(max_length=128, blank=False, unique=True)
    # study_configuration = models.TextField(blank=False)
    # algorithm = models.CharField(max_length=128, blank=False)

    # status = models.CharField(max_length=128, blank=False)
    # created_time = models.DateTimeField(auto_now_add=True)
    # updated_time = models.DateTimeField(auto_now=True)

    def __str__(self):
        return "{}-{}".format(self.id, self.name)

    @classmethod
    def create(
        cls,
        name,
        study_configuration,
        algorithm="RandomSearchAlgorithm",
        status="Pending",
    ):
        study = cls()
        study.name = name
        study.study_configuration = study_configuration
        study.algorithm = algorithm
        study.status = status
        # study.save()
        return study

    def to_json(self):
        return {
            "id": self.id,
            "name": self.name,
            "study_configuration": self.study_configuration,
            "algorithm": self.algorithm,
            "status": self.status
            # "created_time": self.created_time,
            # "updated_time": self.updated_time
        }


class Trial:
    # TODO: Use foreign key or not
    # study_name = models.ForeignKey(Study, related_name="trial_study", to_field=Study.name)
    # study_name = models.CharField(max_length=128, blank=False)
    # name = models.CharField(max_length=128, blank=False)
    # parameter_values = models.TextField(blank=True, null=True)
    # objective_value = models.FloatField(blank=True, null=True)

    # status = models.CharField(max_length=128, blank=False)
    # created_time = models.DateTimeField(auto_now_add=True)
    # updated_time = models.DateTimeField(auto_now=True)

    def __str__(self):
        return "{}-{}".format(self.id, self.name)

    @classmethod
    def create(cls, study_name, name):
        trial = cls()
        trial.study_name = study_name
        trial.name = name
        trial.status = "Pending"
        trial.trainer = None
        # trial.save()
        return trial

    def to_json(self):
        return {
            "id": self.id,
            "study_name": self.study_name,
            "name": self.name,
            "parameter_values": self.parameter_values,
            "objective_value": self.objective_value,
            "status": self.status
            # "created_time": self.created_time,
            # "updated_time": self.updated_time
        }


class TrialMetric:
    # trial_id = models.IntegerField(blank=False)
    # training_step = models.IntegerField(blank=True, null=True)
    # objective_value = models.FloatField(blank=True, null=True)

    # created_time = models.DateTimeField(auto_now_add=True)
    # updated_time = models.DateTimeField(auto_now=True)

    def __str__(self):
        return "Id: {}, trial id: {}, training_step: {}".format(
            self.id, self.trial_id, self.training_step
        )

    @classmethod
    def create(cls, trial_id, training_step, objective_value):
        trial_metric = cls()
        trial_metric.trial_id = trial_id
        trial_metric.training_step = training_step
        trial_metric.objective_value = objective_value
        trial_metric.save()
        return trial_metric

    def to_json(self):
        return {
            "id": self.id,
            "trial_id": self.trial_id,
            "training_step": self.training_step,
            "objective_value": self.objective_value
            # "created_time": self.created_time,
            # "updated_time": self.updated_time
        }


class Algorithm:
    # name = models.CharField(max_length=128, blank=False)

    # status = models.CharField(max_length=128, blank=False)
    ##created_time = models.DateTimeField(auto_now_add=True)
    # updated_time = models.DateTimeField(auto_now=True)

    def __str__(self):
        return "{}-{}".format(self.id, self.name)

    @classmethod
    def create(cls, name):
        algorithm = cls()
        algorithm.name = name
        algorithm.status = "AVAIABLE"
        algorithm.save()
        return algorithm

    def to_json(self):
        return {"name": self.name}



import random
import os
import sys
import collections
import itertools
import pickle as pkl
import scipy.sparse as sp

import networkx as nx
import numpy as np
from sklearn.cluster import SpectralClustering
from sklearn import gaussian_process
from scipy.optimize import minimize
from bayes_opt import BayesianOptimization


def rand(size, a, b, decimals=4):
    res = np.random.random_sample(size) * (b - a) + a
    if decimals is not None:
        return np.around(res, decimals=decimals)
    return res


def softmax(x):
    e_x = np.exp(x - np.max(x))
    return e_x / e_x.sum()


def split_network(G, N):
    sc = SpectralClustering(N, affinity="precomputed")
    return sc.fit_predict(nx.adjacency_matrix(G))


def random_walk_induced_graph_sampling(G, N, T=100, growth_size=2, n_starts=5):
    # Refer to https://github.com/Ashish7129/Graph-Sampling
    G = nx.convert_node_labels_to_integers(G, 0, "default", True)
    for n, data in G.nodes(data=True):
        G.node[n]["id"] = n
    n_node = G.number_of_nodes()
    labels = nx.get_node_attributes(G, "label")
    candidate_set = set()
    if len(labels) > 0:
        candidate_set_value = set()
        l = np.random.permutation(list(labels.keys()))
        for i in l:
            for j in labels[i]:
                if j not in candidate_set_value:
                    candidate_set.add(i)
                    candidate_set_value.add(j)
    while len(candidate_set) < n_starts:
        candidate_set.add(np.random.random_integers(n_node))
    candidate_set = list(candidate_set)[:n_starts]
    sampled_nodes = set()
    for i, temp_node in enumerate(candidate_set):
        sampled_nodes.add(G.node[temp_node]["id"])
        iter_ = 1
        nodes_before_t_iter = 0
        curr_node = temp_node
        N_t = int(N / n_starts) * (i + 1)
        while len(sampled_nodes) < N_t:
            edges = [n for n in G.neighbors(curr_node)]
            index_of_edge = random.randint(0, len(edges) - 1)
            chosen_node = edges[index_of_edge]
            sampled_nodes.add(G.node[chosen_node]["id"])
            curr_node = chosen_node
            iter_ += 1
            if iter_ % T == 0:
                if len(sampled_nodes) - nodes_before_t_iter < growth_size:
                    curr_node = random.randint(0, n_node - 1)
                nodes_before_t_iter = len(sampled_nodes)
    sampled_graph = G.subgraph(sampled_nodes)
    return sampled_graph


def generate_mask(dataset_path, radio=0.8):
    G = load_graph(
        os.path.join(dataset_path, "graph.edgelist"),
        os.path.join(dataset_path, "label.txt"),
    )
    labels = nx.get_node_attributes(G, "label")
    l = np.random.permutation(list(labels.keys()))
    n = int(radio * len(l))
    with open(os.path.join(dataset_path, "label_mask_train"), "w") as f:
        for i in l[:n]:
            print(i, file=f)
    with open(os.path.join(dataset_path, "label_mask_test"), "w") as f:
        for i in l[n:]:
            print(i, file=f)


def write_with_create(path):
    dirpath = os.path.dirname(path)
    if not os.path.exists(dirpath):
        os.makedirs(dirpath)
    return open(path, "w")


def load_graph(edgelist_filename, label_name=None):
    G = nx.read_edgelist(edgelist_filename, nodetype=int)
    if label_name is not None:
        labels = np.loadtxt(label_name, dtype=int)
        # multi-label
        l = collections.defaultdict(list)
        for i, j in labels:
            l[i].append(j)
        # Warning:: The call order of arguments `values` and `name` switched between v1.x & v2.x.
        nx.set_node_attributes(G, l, "label")
    print("load graph", G.number_of_nodes(), G.number_of_edges())
    return G


def run_target_model(
    method, input_filename, output_dir, embedding_test_dir, debug=True, **kargs
):
    sys.path.append(embedding_test_dir)
    from src.baseline import baseline

    with cd(embedding_test_dir):
        baseline(
            method,
            None,
            kargs["emd_size"],
            input_filename,
            output_dir,
            debug=debug,
            **kargs
        )


def run_test(task, dataset_name, models, labels, save_filename, embedding_test_dir):
    sys.path.append(embedding_test_dir)
    from src.test import test

    args = {}
    if task == "classification":
        args["radio"] = [0.8]
        args["label_name"] = labels
        evalution = None
    elif task == "link_predict":
        evalution = "AUC"
        args["data_dir"] = labels
    args["sampling_mapping"] = {"Flickr": 100000, "wiki": 1000000}

    with cd(embedding_test_dir):
        test(task, evalution, dataset_name, models, save_filename=save_filename, **args)


def get_names(method, **args):
    kargs = args
    if method == "node2vec":
        embedding_filename = os.path.join(
            "{}_{:d}_{:d}_{:d}_{:d}_{:.4f}_{:.4f}".format(
                method,
                kargs["emd_size"],
                kargs["num-walks"],
                kargs["walk-length"],
                kargs["window-size"],
                kargs["p"],
                kargs["q"],
            )
        )
    elif method == "deepwalk":
        embedding_filename = os.path.join(
            "{}_{:d}_{:d}_{:d}_{:d}".format(
                method,
                kargs["emd_size"],
                kargs["number-walks"],
                kargs["walk-length"],
                kargs["window-size"],
            )
        )
    elif method == "gcn":
        # embedding_filename = os.path.join("{}_{:d}_{:d}_{:.4f}".format(method, kargs['epochs'], kargs['hidden1'], kargs['learning_rate']))
        embedding_filename = os.path.join(
            "{}_{:d}_{:d}_{:.4f}_{:.4f}_{:.4f}".format(
                method,
                kargs["epochs"],
                kargs["hidden1"],
                kargs["learning_rate"],
                kargs["dropout"],
                kargs["weight_decay"],
            )
        )
    elif method == "AROPE":
        embedding_filename = os.path.join(
            "{}_{}_".format(method, kargs["emd_size"])
            + "_".join(
                [
                    "{:.4f}".format(kargs["w{}".format(i + 1)])
                    for i in range(kargs["order"])
                ]
            )
        )
    return embedding_filename


def random_with_bound_type(bound, type_):
    res = []
    for b, t in zip(bound, type_):
        if t == int:
            res.append(random.randint(*b))
        elif t == float:
            res.append(rand(1, *b)[0])
        else:
            res.append(None)
    return res


def find_b_opt_max(gp, p_bound, p_type, w=None, n_warmup=100000, n_iter=100):
    """
    refer to acq_max https://github.com/fmfn/BayesianOptimization/blob/master/bayes_opt/util.py
    """
    X = []
    for k in range(n_warmup):
        X.append(random_with_bound_type(p_bound, p_type))
    if w is not None:
        X = np.hstack((X, np.tile(w, (len(X), 1))))
    y = gp.predict(X)
    ind = np.argmax(y)
    x_max, y_max = X[ind], y[ind]
    temp_w = [] if w is None else w

    def utility(x, kappa=2.576):
        mean, std = gp.predict([list(x) + list(temp_w)], return_std=True)
        # print("######### mean, std ", x, mean, std)
        return (mean + kappa * std)[0]

    for i in range(n_iter):
        x_try = random_with_bound_type(p_bound, p_type)
        res = minimize(lambda x: -utility(x), x_try, bounds=p_bound, method="L-BFGS-B")
        if not res.success:
            continue
        if -res.fun >= y_max:
            x_max = res.x
            y_max = -res.fun

    return x_max, y_max


class cd:
    """Context manager for changing the current working directory"""

    def __init__(self, newPath):
        self.newPath = os.path.expanduser(newPath)

    def __enter__(self):
        self.savedPath = os.getcwd()
        os.chdir(self.newPath)

    def __exit__(self, etype, value, traceback):
        os.chdir(self.savedPath)


class K(gaussian_process.kernels.Kernel):
    def __init__(self, n=3):
        self.n = n
        self.kernels = [
            gaussian_process.kernels.Matern(nu=2.5),
            gaussian_process.kernels.Matern(nu=2.5),
        ]

    def __call__(self, X, Y=None):
        n = self.n
        if Y is None:
            Y = X
        return self.kernels[0](X[:, :n], Y[:, :n]) * self.kernels[1](X[:, n:], Y[:, n:])

    def diag(self, X):
        n = self.n
        return self.kernels[0].diag(X[:, :n]) * self.kernels[1].diag(X[:, n:])

    def is_stationary(self):
        return np.all([kernel.is_stationary() for kernel in self.kernels])


class GaussianProcessRegressor(object):
    def __init__(self, kernel=None):
        if kernel is None:
            kernel = gaussian_process.kernels.Matern(nu=2.5)
        self.gp = gaussian_process.GaussianProcessRegressor(
            kernel=kernel, alpha=1e-6, normalize_y=True, n_restarts_optimizer=10
        )

    def fit(self, X, y):
        self.gp.fit(X, y)

    def predict(self, p_bound, type_, w=None):
        return find_b_opt_max(self.gp, p_bound, type_, w)


class meta_learner(BayesianOptimization):
    def set_kernel(kernel):
        self._gp.kernel = kernel


class RandomState(object):
    def __init__(self):
        self.state = None

    def set_seed(self, seed):
        random.seed(seed)
        np.random.seed(seed)

    def save_state(self):
        self.state = (random.getstate(), np.random.get_state())

    def load_state(self):
        random.setstate(self.state[0])
        np.random.set_state(self.state[1])


class Params(object):
    def __init__(self):
        eps = 1e-6

    def set_space(self, config):
        """
        Set up a study for advisor.
        config: [{
            "parameterName": "num_layers",
            "type": "DISCRETE",
            "feasiblePoints": "1,2,3,4",
        },{
            "parameterName": "dropout",
            "type": "DOUBLE",
            "minValue": 0.1,
            "maxValue": 0.9,
            "scalingType": "LINEAR"
        }]"""
        # len = len(config)
        self.arg_name = []
        self.arg_len = []
        self.arg_typ = []
        # len = sum(arg_len)
        self.type_ = []
        self.bound = []
        self.config = {}
        self.id2cate = {}

        for para in config:
            if para["type"] == "DOUBLE":
                self.arg_name.append(para["parameterName"])
                self.arg_len.append(1)
                self.arg_typ.append("value")
                self.type_.append(float)
                self.bound.append((para["minValue"], para["maxValue"]))
            elif para["type"] == "INTEGER":
                self.arg_name.append(para["parameterName"])
                self.arg_len.append(1)
                self.arg_typ.append("value")
                self.type_.append(int)
                self.bound.append((para["minValue"], para["maxValue"]))
            elif para["type"] == "DISCRETE" or para["type"] == "CATEGORICAL":
                cate_list = para["feasiblePoints"].split(",")
                cate_list = list(map(lambda x: x.strip(), cate_list))
                cate_len = len(cate_list)
                self.config[para["parameterName"]] = cate_list
                self.id2cate[para["parameterName"]] = {
                    x: y for y, x in enumerate(cate_list)
                }

                self.arg_name.append(para["parameterName"])
                self.arg_len.append(cate_len)
                self.arg_typ.append("argmax")
                self.type_.extend([float for i in range(cate_len)])
                self.bound.extend([(0, 1) for i in range(cate_len)])

    def get_type(self, ps=None):
        if ps is None:
            return self.type_
        return [self.type_[self.ind[p]] for p in ps]

    def get_bound(self, ps=None):
        if ps is None:
            return self.bound
        return [self.bound[self.ind[p]] for p in ps]

    def x2dict(self, X):
        # convert X to para(dict)
        types = self.get_type()
        bound = np.array(self.get_bound())

        X = np.clip(X, bound[:, 0], bound[:, 1])
        para = {}
        ind = 0
        for name, length, arg_typ in zip(self.arg_name, self.arg_len, self.arg_typ):
            if arg_typ == "value":
                if types[ind] == int:
                    para[name] = round(X[ind])
                elif types[ind] == float:
                    para[name] = round(X[ind], 4)
            else:
                max_ind = np.argmax(X[ind : ind + length])
                para[name] = self.config[name][max_ind]
            ind += length

        return para

    def dict2x(self, para):
        # convert para(dict) to X
        types = self.get_type()

        def one_hot(ind, leng):
            fin = [0 for i in range(leng)]
            fin[ind] = 1
            return fin

        X_fin = []
        ind = 0
        for name, length, arg_typ in zip(self.arg_name, self.arg_len, self.arg_typ):
            if arg_typ == "value":
                X_fin.append(para[name])
            else:
                max_ind = self.id2cate[name][para[name]]
                X_fin.extend(one_hot(max_ind, length))
            ind += length

        return X_fin

    def random_x(self):
        # sample a random X and give it to convert
        type_ = self.get_type()
        bound = self.get_bound()
        res = random_with_bound_type(bound, type_)
        return res


def analysis_result(data_dir):
    fs = os.listdir(data_dir)
    fs = np.array(
        [np.loadtxt(os.path.join(data_dir, f)) for f in fs if not f.endswith("names")]
    )
    print(fs.shape)
    scale = 100
    d = (fs[:, 0] * scale).astype(int)
    for k, v in collections.Counter(d).most_common():
        print(k * 1.0 / scale, v, "{:.2f}".format(v * 1.0 / fs.shape[0]))


def check_label(data_dir):
    fn = os.path.join(data_dir, "label.txt")
    d = np.loadtxt(fn, dtype=int)
    c = collections.Counter(d[:, 1])
    for k, v in c.most_common():
        print(k, v)


def convert_gcn_data(dataset_name, input_dir, output_dir):
    def parse_index_file(filename):
        """Parse index file."""
        index = []
        for line in open(filename):
            index.append(int(line.strip()))
        return index

    names = ["x", "y", "tx", "ty", "allx", "ally", "graph"]
    objects = []
    for i in range(len(names)):
        with open(
            os.path.join(input_dir, "ind.{}.{}".format(dataset_name, names[i])), "rb"
        ) as f:
            if sys.version_info > (3, 0):
                objects.append(pkl.load(f, encoding="latin1"))
            else:
                objects.append(pkl.load(f))

    x, y, tx, ty, allx, ally, graph = tuple(objects)
    test_idx_reorder = parse_index_file(
        os.path.join(input_dir, "ind.{}.test.index".format(dataset_name))
    )
    test_idx_range = np.sort(test_idx_reorder)

    if dataset_name == "citeseer":
        # Fix citeseer dataset (there are some isolated nodes in the graph)
        # Find isolated nodes, add them as zero-vecs into the right position
        test_idx_range_full = range(min(test_idx_reorder), max(test_idx_reorder) + 1)
        tx_extended = sp.lil_matrix((len(test_idx_range_full), x.shape[1]))
        tx_extended[test_idx_range - min(test_idx_range), :] = tx
        tx = tx_extended
        ty_extended = np.zeros((len(test_idx_range_full), y.shape[1]))
        ty_extended[test_idx_range - min(test_idx_range), :] = ty
        ty = ty_extended

    features = sp.vstack((allx, tx)).tolil()
    features[test_idx_reorder, :] = features[test_idx_range, :]
    G = nx.from_dict_of_lists(graph)
    adj = nx.adjacency_matrix(G)

    labels = np.vstack((ally, ty))
    labels[test_idx_reorder, :] = labels[test_idx_range, :]
    print(G.number_of_nodes(), G.number_of_edges())
    with open("data/{}/label.txt".format(dataset_name), "w") as f:
        for i in range(len(labels)):
            assert sum(labels[i]) > 0
            for j, k in enumerate(labels[i]):
                if k == 1:
                    print(i, j, file=f)

    with open("data/{}/graph.edgelist".format(dataset_name), "w") as f:
        for i, j in G.edges():
            print(i, j, file=f)

    sp.save_npz("data/{}/features.npz".format(dataset_name), features.tocsr())


if __name__ == "__main__":
    # analysis_result('result/BlogCatalog/cf/')
    # check_label('data/citeseer/sampled/s1/')
    for i in range(5):
        generate_mask("data/pubmed/sampled/s{}".format(i))
    # convert_gcn_data('pubmed', 'embedding_test/src/baseline/gcn/gcn/data/', 'data/')
